#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Functionality related to allocations."""

import random

from ironic_lib import metrics_utils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
import tenacity

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.conductor import task_manager
from ironic import objects


CONF = cfg.CONF
LOG = log.getLogger(__name__)
METRICS = metrics_utils.get_metrics_logger(__name__)


def do_allocate(context, allocation):
    """Process the allocation.

    This call runs in a separate thread on a conductor. It finds suitable
    nodes for the allocation and reserves one of them.

    This call does not raise exceptions since it's designed to work
    asynchronously.

    :param context: an admin context
    :param allocation: an allocation object
    """
    try:
        nodes = _candidate_nodes(context, allocation)
        _allocate_node(context, allocation, nodes)
    except exception.AllocationFailed as exc:
        LOG.error(str(exc))
        _allocation_failed(allocation, exc)
    except Exception as exc:
        LOG.exception("Unexpected exception during processing of "
                      "allocation %s", allocation.uuid)
        reason = _("Unexpected exception during allocation: %s") % exc
        _allocation_failed(allocation, reason)


def verify_node_for_deallocation(node, allocation):
    """Verify that allocation can be removed for the node.

    :param node: a node object
    :param allocation: an allocation object associated with the node
    """
    if node.maintenance:
        # Allocations can always be removed in the maintenance mode.
        return

    if (node.target_provision_state
            and node.provision_state not in states.UPDATE_ALLOWED_STATES):
        msg = (_("Cannot remove allocation %(uuid)s for node %(node)s, "
                 "because the node is in state %(state)s where updates are "
                 "not allowed (and maintenance mode is off)") %
               {'node': node.uuid, 'uuid': allocation.uuid,
                'state': node.provision_state})
        raise exception.InvalidState(msg)

    if node.provision_state == states.ACTIVE:
        msg = (_("Cannot remove allocation %(uuid)s for node %(node)s, "
                 "because the node is active (and maintenance mode is off)") %
               {'node': node.uuid, 'uuid': allocation.uuid})
        raise exception.InvalidState(msg)


def _allocation_failed(allocation, reason):
    """Failure handler for the allocation."""
    try:
        allocation.node_id = None
        allocation.state = states.ERROR
        allocation.last_error = str(reason)
        allocation.save()
    except exception.AllocationNotFound as exc:
        LOG.debug('Not saving a failed allocation: %s', exc)
    except Exception:
        LOG.exception('Could not save the failed allocation %s',
                      allocation.uuid)


def _traits_match(traits, node):
    return {t.trait for t in node.traits.objects}.issuperset(traits)


def _candidate_nodes(context, allocation):
    """Get a list of candidate nodes for the allocation."""
    # NOTE(dtantsur): not checking the retired flag because it's impossible
    # (by the API contract) to have a retired node in the available state.
    filters = {'resource_class': allocation.resource_class,
               'provision_state': states.AVAILABLE,
               'associated': False,
               'with_power_state': True,
               'maintenance': False}
    if allocation.candidate_nodes:
        # NOTE(dtantsur): we assume that candidate_nodes were converted to
        # UUIDs on the API level.
        filters['uuid_in'] = allocation.candidate_nodes
    if allocation.owner:
        filters['project'] = allocation.owner

    nodes = objects.Node.list(context, filters=filters)

    if not nodes:
        if allocation.candidate_nodes:
            error = _("none of the requested nodes are available and match "
                      "the resource class %s") % allocation.resource_class
        else:
            error = _("no available nodes match the resource class %s") % (
                allocation.resource_class)
        raise exception.AllocationFailed(uuid=allocation.uuid, error=error)

    # TODO(dtantsur): database-level filtering?
    if allocation.traits:
        traits = set(allocation.traits)
        nodes = [n for n in nodes if _traits_match(traits, n)]
        if not nodes:
            error = (_("no suitable nodes have the requested traits %s") %
                     ', '.join(traits))
            raise exception.AllocationFailed(uuid=allocation.uuid, error=error)

    # NOTE(dtantsur): make sure that parallel allocations do not try the nodes
    # in the same order.
    random.shuffle(nodes)

    # NOTE(sbaker): if the allocation name matches a node name, attempt that
    # node first. This will reduce confusion when nodes have the same naming
    # scheme as allocations.
    if allocation.name:
        for i, node in enumerate(nodes):
            if node.name == allocation.name:
                nodes.insert(0, nodes.pop(i))
                break

    LOG.debug('%(count)d nodes are candidates for allocation %(uuid)s',
              {'count': len(nodes), 'uuid': allocation.uuid})
    return nodes


def _verify_node(node, allocation):
    """Check that the node still satisfiest the request."""
    if node.maintenance:
        LOG.debug('Node %s is now in maintenance, skipping',
                  node.uuid)
        return False

    if node.instance_uuid:
        LOG.debug('Node %(node)s is already associated with instance '
                  '%(inst)s, skipping',
                  {'node': node.uuid, 'inst': node.instance_uuid})
        return False

    if node.provision_state != states.AVAILABLE:
        LOG.debug('Node %s is no longer available, skipping',
                  node.uuid)
        return False

    if node.resource_class != allocation.resource_class:
        LOG.debug('Resource class of node %(node)s no longer '
                  'matches requested resource class %(rsc)s for '
                  'allocation %(uuid)s, skipping',
                  {'node': node.uuid,
                   'rsc': allocation.resource_class,
                   'uuid': allocation.uuid})
        return False

    if allocation.traits and not _traits_match(set(allocation.traits), node):
        LOG.debug('List of traits of node %(node)s no longer '
                  'matches requested traits %(traits)s for '
                  'allocation %(uuid)s, skipping',
                  {'node': node.uuid,
                   'traits': allocation.traits,
                   'uuid': allocation.uuid})
        return False

    return True


# NOTE(dtantsur): instead of trying to allocate each node
# node_locked_retry_attempt times, we try to allocate *any* node the same
# number of times. This avoids getting stuck on a node reserved e.g. for power
# sync periodic task.
@tenacity.retry(
    retry=tenacity.retry_if_exception_type(exception.AllocationFailed),
    stop=tenacity.stop_after_attempt(
        CONF.conductor.node_locked_retry_attempts),
    wait=tenacity.wait_fixed(
        CONF.conductor.node_locked_retry_interval),
    reraise=True)
def _allocate_node(context, allocation, nodes):
    """Go through the list of nodes and try to allocate one of them."""
    retry_nodes = []
    for node in nodes:
        try:
            # NOTE(dtantsur): retries are done for all nodes above, so disable
            # per-node retry. Also disable loading the driver, since the
            # current conductor may not have the requried hardware type or
            # interfaces (it's picked at random).
            with task_manager.acquire(context, node.uuid, shared=False,
                                      retry=False, load_driver=False,
                                      purpose='allocating') as task:
                # NOTE(dtantsur): double-check the node details, since they
                # could have changed before we acquired the lock.
                if not _verify_node(task.node, allocation):
                    continue

                allocation.node_id = task.node.id
                allocation.state = states.ACTIVE
                # NOTE(dtantsur): the node.instance_uuid and allocation_id are
                # updated inside of the save() call within the same
                # transaction to avoid races. NodeAssociated can be raised if
                # another process allocates this node first.
                allocation.save()
                LOG.info('Node %(node)s has been successfully reserved for '
                         'allocation %(uuid)s',
                         {'node': node.uuid, 'uuid': allocation.uuid})
                return allocation
        except exception.NodeLocked:
            LOG.debug('Node %s is currently locked, moving to the next one',
                      node.uuid)
            retry_nodes.append(node)
        except exception.NodeAssociated:
            LOG.debug('Node %s is already associated, moving to the next one',
                      node.uuid)

    # NOTE(dtantsur): rewrite the passed list to only contain the nodes that
    # are worth retrying. Do not include nodes that are no longer suitable.
    nodes[:] = retry_nodes

    if nodes:
        error = _('could not reserve any of %d suitable nodes') % len(nodes)
    else:
        error = _('all nodes were filtered out during reservation')

    raise exception.AllocationFailed(uuid=allocation.uuid, error=error)


def backfill_allocation(context, allocation, node_id):
    """Assign the previously allocated node to the node allocation.

    This is not the actual allocation process, but merely backfilling of
    allocation_uuid for a previously allocated node.

    :param context: an admin context
    :param allocation: an allocation object associated with the node
    :param node_id: An ID of the node.
    :raises: AllocationFailed if the node does not match the allocation
    :raises: NodeAssociated if the node is already associated with another
        instance or allocation.
    :raises: InstanceAssociated if the allocation's UUID is already used
        on another node as instance_uuid.
    :raises: NodeNotFound if the node with the provided ID cannot be found.
    """
    try:
        _do_backfill_allocation(context, allocation, node_id)
    except (exception.AllocationFailed,
            exception.InstanceAssociated,
            exception.NodeAssociated,
            exception.NodeNotFound) as exc:
        with excutils.save_and_reraise_exception():
            LOG.error(str(exc))
            _allocation_failed(allocation, exc)
    except Exception as exc:
        with excutils.save_and_reraise_exception():
            LOG.exception("Unexpected exception during backfilling of "
                          "allocation %s", allocation.uuid)
            reason = _("Unexpected exception during allocation: %s") % exc
            _allocation_failed(allocation, reason)


def _do_backfill_allocation(context, allocation, node_id):
    with task_manager.acquire(context, node_id,
                              purpose='allocation backfilling') as task:
        node = task.node

        errors = []

        # NOTE(dtantsur): this feature is not designed to bypass the allocation
        # mechanism, but to backfill allocations for active nodes, hence this
        # check.
        if node.provision_state != states.ACTIVE:
            errors.append(_('Node must be in the "active" state, but the '
                            'current state is "%s"') % node.provision_state)

        # NOTE(dtantsur): double-check that the node is still suitable.
        if (allocation.resource_class
                and node.resource_class != allocation.resource_class):
            errors.append(_('Resource class %(curr)s does not match '
                            'the requested resource class %(rsc)s')
                          % {'curr': node.resource_class,
                             'rsc': allocation.resource_class})
        if (allocation.traits
                and not _traits_match(set(allocation.traits), node)):
            errors.append(_('List of traits %(curr)s does not match '
                            'the requested traits %(traits)s')
                          % {'curr': node.traits,
                             'traits': allocation.traits})
        if (allocation.candidate_nodes
                and node.uuid not in allocation.candidate_nodes):
            errors.append(_('Candidate nodes must be empty or contain the '
                            'target node, but got %s')
                          % allocation.candidate_nodes)

        if errors:
            error = _('Cannot backfill an allocation for node %(node)s: '
                      '%(errors)s') % {'node': node.uuid,
                                       'errors': '; '.join(errors)}
            raise exception.AllocationFailed(uuid=allocation.uuid, error=error)

        allocation.node_id = task.node.id
        allocation.state = states.ACTIVE
        # NOTE(dtantsur): the node.instance_uuid and allocation_id are
        # updated inside of the save() call within the same
        # transaction to avoid races. NodeAssociated can be raised if
        # another process allocates this node first.
        allocation.save()
        LOG.info('Node %(node)s has been successfully reserved for '
                 'allocation %(uuid)s',
                 {'node': node.uuid, 'uuid': allocation.uuid})



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

COLLECT_INFO = (('wwn', 'WWN'), ('serial', 'SERIAL_SHORT'),
                ('wwn_with_extension', 'WWN_WITH_EXTENSION'),
                ('wwn_vendor_extension', 'WWN_VENDOR_EXTENSION'))


def get_devices_wwn(devices, module):
    try:
        import pyudev
        # NOTE(pas-ha) creating context might fail if udev is missing
        context = pyudev.Context()
    except ImportError:
        module.warn('Can not collect "wwn", "wwn_with_extension", '
                    '"wwn_vendor_extension" and "serial" when using '
                    'root device hints because there\'s no UDEV python '
                    'binds installed')
        return {}

    dev_dict = {}
    for device in devices:
        name = '/dev/' + device
        try:
            udev = pyudev.Devices.from_device_file(context, name)
        except (ValueError, EnvironmentError, pyudev.DeviceNotFoundError) as e:
            module.warn('Device %(dev)s is inaccessible, skipping... '
                        'Error: %(error)s' % {'dev': name, 'error': e})
            continue

        dev_dict[device] = {}
        for key, udev_key in COLLECT_INFO:
            candidate = udev.get('ID_%s' % udev_key)
            if candidate:
                candidate = candidate.lower()
            dev_dict[device][key] = candidate

    return {"ansible_facts": {"devices_wwn": dev_dict}}


def main():
    module = AnsibleModule(  # noqa This is normal for Ansible modules.
        argument_spec=dict(
            devices=dict(required=True, type='list'),
        ),
        supports_check_mode=True,
    )

    devices = module.params['devices']
    data = get_devices_wwn(devices, module)
    module.exit_json(**data)


from ansible.module_utils.basic import *  # noqa
if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from ironic.common import context
from ironic.db import api as dbapi
from ironic import objects
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils


class TestDeployTemplateObject(db_base.DbTestCase, obj_utils.SchemasTestMixIn):

    def setUp(self):
        super(TestDeployTemplateObject, self).setUp()
        self.ctxt = context.get_admin_context()
        self.fake_template = db_utils.get_test_deploy_template()

    @mock.patch.object(dbapi.IMPL, 'create_deploy_template', autospec=True)
    def test_create(self, mock_create):
        template = objects.DeployTemplate(context=self.context,
                                          **self.fake_template)

        mock_create.return_value = db_utils.get_test_deploy_template()

        template.create()

        args, _kwargs = mock_create.call_args
        self.assertEqual(1, mock_create.call_count)

        self.assertEqual(self.fake_template['name'], template.name)
        self.assertEqual(self.fake_template['steps'], template.steps)
        self.assertEqual(self.fake_template['extra'], template.extra)

    @mock.patch.object(dbapi.IMPL, 'update_deploy_template', autospec=True)
    def test_save(self, mock_update):
        template = objects.DeployTemplate(context=self.context,
                                          **self.fake_template)
        template.obj_reset_changes()

        mock_update.return_value = db_utils.get_test_deploy_template(
            name='CUSTOM_DT2')

        template.name = 'CUSTOM_DT2'
        template.save()

        mock_update.assert_called_once_with(
            self.fake_template['uuid'],
            {'name': 'CUSTOM_DT2', 'version': objects.DeployTemplate.VERSION})

        self.assertEqual('CUSTOM_DT2', template.name)

    @mock.patch.object(dbapi.IMPL, 'destroy_deploy_template', autospec=True)
    def test_destroy(self, mock_destroy):
        template = objects.DeployTemplate(context=self.context,
                                          id=self.fake_template['id'])

        template.destroy()

        mock_destroy.assert_called_once_with(self.fake_template['id'])

    @mock.patch.object(dbapi.IMPL, 'get_deploy_template_by_id', autospec=True)
    def test_get_by_id(self, mock_get):
        mock_get.return_value = self.fake_template

        template = objects.DeployTemplate.get_by_id(
            self.context, self.fake_template['id'])

        mock_get.assert_called_once_with(self.fake_template['id'])
        self.assertEqual(self.fake_template['name'], template.name)
        self.assertEqual(self.fake_template['uuid'], template.uuid)
        self.assertEqual(self.fake_template['steps'], template.steps)
        self.assertEqual(self.fake_template['extra'], template.extra)

    @mock.patch.object(dbapi.IMPL, 'get_deploy_template_by_uuid',
                       autospec=True)
    def test_get_by_uuid(self, mock_get):
        mock_get.return_value = self.fake_template

        template = objects.DeployTemplate.get_by_uuid(
            self.context, self.fake_template['uuid'])

        mock_get.assert_called_once_with(self.fake_template['uuid'])
        self.assertEqual(self.fake_template['name'], template.name)
        self.assertEqual(self.fake_template['uuid'], template.uuid)
        self.assertEqual(self.fake_template['steps'], template.steps)
        self.assertEqual(self.fake_template['extra'], template.extra)

    @mock.patch.object(dbapi.IMPL, 'get_deploy_template_by_name',
                       autospec=True)
    def test_get_by_name(self, mock_get):
        mock_get.return_value = self.fake_template

        template = objects.DeployTemplate.get_by_name(
            self.context, self.fake_template['name'])

        mock_get.assert_called_once_with(self.fake_template['name'])
        self.assertEqual(self.fake_template['name'], template.name)
        self.assertEqual(self.fake_template['uuid'], template.uuid)
        self.assertEqual(self.fake_template['steps'], template.steps)
        self.assertEqual(self.fake_template['extra'], template.extra)

    @mock.patch.object(dbapi.IMPL, 'get_deploy_template_list', autospec=True)
    def test_list(self, mock_list):
        mock_list.return_value = [self.fake_template]

        templates = objects.DeployTemplate.list(self.context)

        mock_list.assert_called_once_with(limit=None, marker=None,
                                          sort_dir=None, sort_key=None)
        self.assertEqual(1, len(templates))
        self.assertEqual(self.fake_template['name'], templates[0].name)
        self.assertEqual(self.fake_template['uuid'], templates[0].uuid)
        self.assertEqual(self.fake_template['steps'], templates[0].steps)
        self.assertEqual(self.fake_template['extra'], templates[0].extra)

    @mock.patch.object(dbapi.IMPL, 'get_deploy_template_list_by_names',
                       autospec=True)
    def test_list_by_names(self, mock_list):
        mock_list.return_value = [self.fake_template]

        names = [self.fake_template['name']]
        templates = objects.DeployTemplate.list_by_names(self.context, names)

        mock_list.assert_called_once_with(names)
        self.assertEqual(1, len(templates))
        self.assertEqual(self.fake_template['name'], templates[0].name)
        self.assertEqual(self.fake_template['uuid'], templates[0].uuid)
        self.assertEqual(self.fake_template['steps'], templates[0].steps)
        self.assertEqual(self.fake_template['extra'], templates[0].extra)

    @mock.patch.object(dbapi.IMPL, 'get_deploy_template_by_uuid',
                       autospec=True)
    def test_refresh(self, mock_get):
        uuid = self.fake_template['uuid']
        mock_get.side_effect = [dict(self.fake_template),
                                dict(self.fake_template, name='CUSTOM_DT2')]

        template = objects.DeployTemplate.get_by_uuid(self.context, uuid)

        self.assertEqual(self.fake_template['name'], template.name)

        template.refresh()

        self.assertEqual('CUSTOM_DT2', template.name)
        expected = [mock.call(uuid), mock.call(uuid)]
        self.assertEqual(expected, mock_get.call_args_list)
        self.assertEqual(self.context, template._context)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest import mock

from ironic_lib import utils as irlib_utils
from oslo_concurrency import processutils

from ironic.common import exception
from ironic.common import states
from ironic.common import utils as com_utils
from ironic.conductor import steps
from ironic.conductor import task_manager
from ironic.conductor import utils
from ironic.drivers.modules.ansible import deploy as ansible_deploy
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules import fake
from ironic.drivers.modules.network import flat as flat_network
from ironic.drivers.modules import pxe
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as object_utils

INSTANCE_INFO = {
    'image_source': 'fake-image',
    'image_url': 'http://image',
    'image_checksum': 'checksum',
    'image_disk_format': 'qcow2',
    'root_mb': 5120,
    'swap_mb': 0,
    'ephemeral_mb': 0
}

DRIVER_INFO = {
    'deploy_kernel': 'glance://deploy_kernel_uuid',
    'deploy_ramdisk': 'glance://deploy_ramdisk_uuid',
    'ansible_username': 'test',
    'ansible_key_file': '/path/key',
    'ipmi_address': '127.0.0.1',
}
DRIVER_INTERNAL_INFO = {
    'is_whole_disk_image': True,
    'clean_steps': []
}


class AnsibleDeployTestCaseBase(db_base.DbTestCase):

    def setUp(self):
        super(AnsibleDeployTestCaseBase, self).setUp()

        self.config(enabled_hardware_types=['manual-management'],
                    enabled_deploy_interfaces=['ansible'],
                    enabled_power_interfaces=['fake'],
                    enabled_management_interfaces=['fake'])
        node = {
            'driver': 'manual-management',
            'instance_info': INSTANCE_INFO,
            'driver_info': DRIVER_INFO,
            'driver_internal_info': DRIVER_INTERNAL_INFO,
        }
        self.node = object_utils.create_test_node(self.context, **node)


class TestAnsibleMethods(AnsibleDeployTestCaseBase):

    def test__parse_ansible_driver_info(self):
        self.node.driver_info['ansible_deploy_playbook'] = 'spam.yaml'
        playbook, user, key = ansible_deploy._parse_ansible_driver_info(
            self.node, 'deploy')
        self.assertEqual('spam.yaml', playbook)
        self.assertEqual('test', user)
        self.assertEqual('/path/key', key)

    def test__parse_ansible_driver_info_defaults(self):
        self.node.driver_info.pop('ansible_username')
        self.node.driver_info.pop('ansible_key_file')
        self.config(group='ansible',
                    default_username='spam',
                    default_key_file='/ham/eggs',
                    default_deploy_playbook='parrot.yaml')
        playbook, user, key = ansible_deploy._parse_ansible_driver_info(
            self.node, 'deploy')
        # testing absolute path to the playbook
        self.assertEqual('parrot.yaml', playbook)
        self.assertEqual('spam', user)
        self.assertEqual('/ham/eggs', key)

    def test__parse_ansible_driver_info_no_playbook(self):
        self.assertRaises(exception.IronicException,
                          ansible_deploy._parse_ansible_driver_info,
                          self.node, 'test')

    def test__get_node_ip(self):
        di_info = self.node.driver_internal_info
        di_info['agent_url'] = 'http://1.2.3.4:5678'
        self.node.driver_internal_info = di_info
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual('1.2.3.4',
                             ansible_deploy._get_node_ip(task))

    @mock.patch.object(com_utils, 'execute', return_value=('out', 'err'),
                       autospec=True)
    def test__run_playbook(self, execute_mock):
        self.config(group='ansible', playbooks_path='/path/to/playbooks')
        self.config(group='ansible', config_file_path='/path/to/config')
        self.config(group='ansible', verbosity=3)
        self.config(group='ansible', ansible_extra_args='--timeout=100')
        extra_vars = {'foo': 'bar'}

        ansible_deploy._run_playbook(self.node, 'deploy',
                                     extra_vars, '/path/to/key',
                                     tags=['spam'], notags=['ham'])

        execute_mock.assert_called_once_with(
            'env', 'ANSIBLE_CONFIG=/path/to/config',
            'ansible-playbook', '/path/to/playbooks/deploy', '-i',
            '/path/to/playbooks/inventory', '-e', '{"ironic": {"foo": "bar"}}',
            '--tags=spam', '--skip-tags=ham',
            '--private-key=/path/to/key', '-vvv', '--timeout=100')

    @mock.patch.object(com_utils, 'execute', return_value=('out', 'err'),
                       autospec=True)
    def test__run_playbook_default_verbosity_nodebug(self, execute_mock):
        self.config(group='ansible', playbooks_path='/path/to/playbooks')
        self.config(group='ansible', config_file_path='/path/to/config')
        self.config(debug=False)
        extra_vars = {'foo': 'bar'}

        ansible_deploy._run_playbook(self.node, 'deploy', extra_vars,
                                     '/path/to/key')

        execute_mock.assert_called_once_with(
            'env', 'ANSIBLE_CONFIG=/path/to/config',
            'ansible-playbook', '/path/to/playbooks/deploy', '-i',
            '/path/to/playbooks/inventory', '-e', '{"ironic": {"foo": "bar"}}',
            '--private-key=/path/to/key')

    @mock.patch.object(com_utils, 'execute', return_value=('out', 'err'),
                       autospec=True)
    def test__run_playbook_default_verbosity_debug(self, execute_mock):
        self.config(group='ansible', playbooks_path='/path/to/playbooks')
        self.config(group='ansible', config_file_path='/path/to/config')
        self.config(debug=True)
        extra_vars = {'foo': 'bar'}

        ansible_deploy._run_playbook(self.node, 'deploy', extra_vars,
                                     '/path/to/key')

        execute_mock.assert_called_once_with(
            'env', 'ANSIBLE_CONFIG=/path/to/config',
            'ansible-playbook', '/path/to/playbooks/deploy', '-i',
            '/path/to/playbooks/inventory', '-e', '{"ironic": {"foo": "bar"}}',
            '--private-key=/path/to/key', '-vvvv')

    @mock.patch.object(com_utils, 'execute', return_value=('out', 'err'),
                       autospec=True)
    def test__run_playbook_ansible_interpreter_python3(self, execute_mock):
        self.config(group='ansible', playbooks_path='/path/to/playbooks')
        self.config(group='ansible', config_file_path='/path/to/config')
        self.config(group='ansible', verbosity=3)
        self.config(group='ansible',
                    default_python_interpreter='/usr/bin/python3')
        self.config(group='ansible', ansible_extra_args='--timeout=100')
        extra_vars = {'foo': 'bar'}

        ansible_deploy._run_playbook(self.node, 'deploy',
                                     extra_vars, '/path/to/key',
                                     tags=['spam'], notags=['ham'])

        execute_mock.assert_called_once_with(
            'env', 'ANSIBLE_CONFIG=/path/to/config',
            'ansible-playbook', '/path/to/playbooks/deploy', '-i',
            '/path/to/playbooks/inventory', '-e',
            mock.ANY, '--tags=spam', '--skip-tags=ham',
            '--private-key=/path/to/key', '-vvv', '--timeout=100')

        all_vars = execute_mock.call_args[0][7]
        self.assertEqual({"ansible_python_interpreter": "/usr/bin/python3",
                          "ironic": {"foo": "bar"}},
                         json.loads(all_vars))

    @mock.patch.object(com_utils, 'execute', return_value=('out', 'err'),
                       autospec=True)
    def test__run_playbook_ansible_interpreter_override(self, execute_mock):
        self.config(group='ansible', playbooks_path='/path/to/playbooks')
        self.config(group='ansible', config_file_path='/path/to/config')
        self.config(group='ansible', verbosity=3)
        self.config(group='ansible',
                    default_python_interpreter='/usr/bin/python3')
        self.config(group='ansible', ansible_extra_args='--timeout=100')
        self.node.driver_info['ansible_python_interpreter'] = (
            '/usr/bin/python4')
        extra_vars = {'foo': 'bar'}

        ansible_deploy._run_playbook(self.node, 'deploy',
                                     extra_vars, '/path/to/key',
                                     tags=['spam'], notags=['ham'])

        execute_mock.assert_called_once_with(
            'env', 'ANSIBLE_CONFIG=/path/to/config',
            'ansible-playbook', '/path/to/playbooks/deploy', '-i',
            '/path/to/playbooks/inventory', '-e',
            mock.ANY, '--tags=spam', '--skip-tags=ham',
            '--private-key=/path/to/key', '-vvv', '--timeout=100')

        all_vars = execute_mock.call_args[0][7]
        self.assertEqual({"ansible_python_interpreter": "/usr/bin/python4",
                          "ironic": {"foo": "bar"}},
                         json.loads(all_vars))

    @mock.patch.object(com_utils, 'execute',
                       side_effect=processutils.ProcessExecutionError(
                           description='VIKINGS!'),
                       autospec=True)
    def test__run_playbook_fail(self, execute_mock):
        self.config(group='ansible', playbooks_path='/path/to/playbooks')
        self.config(group='ansible', config_file_path='/path/to/config')
        self.config(debug=False)
        extra_vars = {'foo': 'bar'}

        exc = self.assertRaises(exception.InstanceDeployFailure,
                                ansible_deploy._run_playbook,
                                self.node, 'deploy', extra_vars,
                                '/path/to/key')
        self.assertIn('VIKINGS!', str(exc))
        execute_mock.assert_called_once_with(
            'env', 'ANSIBLE_CONFIG=/path/to/config',
            'ansible-playbook', '/path/to/playbooks/deploy', '-i',
            '/path/to/playbooks/inventory', '-e', '{"ironic": {"foo": "bar"}}',
            '--private-key=/path/to/key')

    def test__parse_partitioning_info_root_msdos(self):
        self.config(default_boot_mode='bios', group='deploy')
        expected_info = {
            'partition_info': {
                'label': 'msdos',
                'partitions': {
                    'root':
                        {'number': 1,
                         'part_start': '1MiB',
                         'part_end': '5121MiB',
                         'flags': ['boot']}
                }}}

        i_info = ansible_deploy._parse_partitioning_info(self.node)

        self.assertEqual(expected_info, i_info)

    def test__parse_partitioning_info_all_gpt(self):
        in_info = dict(INSTANCE_INFO)
        in_info['swap_mb'] = 128
        in_info['ephemeral_mb'] = 256
        in_info['ephemeral_format'] = 'ext4'
        in_info['preserve_ephemeral'] = True
        in_info['configdrive'] = 'some-fake-user-data'
        in_info['capabilities'] = {'disk_label': 'gpt'}
        self.node.instance_info = in_info
        self.node.save()

        expected_info = {
            'partition_info': {
                'label': 'gpt',
                'ephemeral_format': 'ext4',
                'preserve_ephemeral': 'yes',
                'partitions': {
                    'bios':
                        {'number': 1,
                         'name': 'bios',
                         'part_start': '1MiB',
                         'part_end': '2MiB',
                         'flags': ['bios_grub']},
                    'ephemeral':
                        {'number': 2,
                         'part_start': '2MiB',
                         'part_end': '258MiB',
                         'name': 'ephemeral'},
                    'swap':
                        {'number': 3,
                         'part_start': '258MiB',
                         'part_end': '386MiB',
                         'name': 'swap'},
                    'configdrive':
                        {'number': 4,
                         'part_start': '386MiB',
                         'part_end': '450MiB',
                         'name': 'configdrive'},
                    'root':
                        {'number': 5,
                         'part_start': '450MiB',
                         'part_end': '5570MiB',
                         'name': 'root'}
                }}}

        i_info = ansible_deploy._parse_partitioning_info(self.node)

        self.assertEqual(expected_info, i_info)

    @mock.patch.object(ansible_deploy.images, 'download_size', autospec=True)
    def test__calculate_memory_req(self, image_mock):
        self.config(group='ansible', extra_memory=1)
        image_mock.return_value = 2000000  # < 2MiB

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual(2, ansible_deploy._calculate_memory_req(task))
            image_mock.assert_called_once_with(task.context, 'fake-image')

    def test__get_python_interpreter(self):
        self.config(group='ansible',
                    default_python_interpreter='/usr/bin/python3')
        self.node.driver_info['ansible_python_interpreter'] = (
            '/usr/bin/python4')

        python_interpreter = ansible_deploy._get_python_interpreter(self.node)

        self.assertEqual('/usr/bin/python4', python_interpreter)

    def test__get_configdrive_path(self):
        self.config(tempdir='/path/to/tmpdir')
        self.assertEqual('/path/to/tmpdir/spam.cndrive',
                         ansible_deploy._get_configdrive_path('spam'))

    def test__prepare_extra_vars(self):
        host_list = [('fake-uuid', '1.2.3.4', 'spam', 'ham'),
                     ('other-uuid', '5.6.7.8', 'eggs', 'vikings')]
        ansible_vars = {"foo": "bar"}
        self.assertEqual(
            {"nodes": [
                {"name": "fake-uuid", "ip": '1.2.3.4',
                 "user": "spam", "extra": "ham"},
                {"name": "other-uuid", "ip": '5.6.7.8',
                 "user": "eggs", "extra": "vikings"}],
                "foo": "bar"},
            ansible_deploy._prepare_extra_vars(host_list, ansible_vars))

    def test__parse_root_device_hints(self):
        hints = {"wwn": "fake wwn", "size": "12345", "rotational": True,
                 "serial": "HELLO"}
        expected = {"wwn": "fake wwn", "size": 12345, "rotational": True,
                    "serial": "hello"}
        props = self.node.properties
        props['root_device'] = hints
        self.node.properties = props
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual(
                expected, ansible_deploy._parse_root_device_hints(task.node))

    def test__parse_root_device_hints_iinfo(self):
        hints = {"wwn": "fake wwn", "size": "12345", "rotational": True,
                 "serial": "HELLO"}
        expected = {"wwn": "fake wwn", "size": 12345, "rotational": True,
                    "serial": "hello"}
        iinfo = self.node.instance_info
        iinfo['root_device'] = hints
        self.node.instance_info = iinfo
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual(
                expected, ansible_deploy._parse_root_device_hints(task.node))

    def test__parse_root_device_hints_override(self):
        hints = {"wwn": "fake wwn", "size": "12345", "rotational": True,
                 "serial": "HELLO"}
        expected = {"wwn": "fake wwn", "size": 12345, "rotational": True,
                    "serial": "hello"}
        props = self.node.properties
        props['root_device'] = {'size': 'no idea'}
        self.node.properties = props
        iinfo = self.node.instance_info
        iinfo['root_device'] = hints
        self.node.instance_info = iinfo
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual(
                expected, ansible_deploy._parse_root_device_hints(task.node))

    def test__parse_root_device_hints_fail_advanced(self):
        hints = {"wwn": "s!= fake wwn",
                 "size": ">= 12345",
                 "name": "<or> spam <or> ham",
                 "rotational": True}
        expected = {"wwn": "s!= fake%20wwn",
                    "name": "<or> spam <or> ham",
                    "size": ">= 12345"}
        props = self.node.properties
        props['root_device'] = hints
        self.node.properties = props
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            exc = self.assertRaises(
                exception.InvalidParameterValue,
                ansible_deploy._parse_root_device_hints, task.node)
            for key, value in expected.items():
                self.assertIn(str(key), str(exc))
                self.assertIn(str(value), str(exc))

    def test__prepare_variables(self):
        i_info = self.node.instance_info
        i_info['image_mem_req'] = 3000
        i_info['image_whatever'] = 'hello'
        self.node.instance_info = i_info
        self.node.save()

        expected = {"image": {"url": "http://image",
                              "validate_certs": "yes",
                              "source": "fake-image",
                              "mem_req": 3000,
                              "disk_format": "qcow2",
                              "checksum": "md5:checksum",
                              "whatever": "hello"}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual(expected,
                             ansible_deploy._prepare_variables(task))

    def test__prepare_variables_root_device_hints(self):
        props = self.node.properties
        props['root_device'] = {"wwn": "fake-wwn"}
        self.node.properties = props
        self.node.save()
        expected = {"image": {"url": "http://image",
                              "validate_certs": "yes",
                              "source": "fake-image",
                              "disk_format": "qcow2",
                              "checksum": "md5:checksum"},
                    "root_device_hints": {"wwn": "fake-wwn"}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual(expected,
                             ansible_deploy._prepare_variables(task))

    def test__prepare_variables_insecure_activated(self):
        self.config(image_store_insecure=True, group='ansible')
        i_info = self.node.instance_info
        i_info['image_checksum'] = 'sha256:checksum'
        self.node.instance_info = i_info
        self.node.save()
        expected = {"image": {"url": "http://image",
                              "validate_certs": "no",
                              "source": "fake-image",
                              "disk_format": "qcow2",
                              "checksum": "sha256:checksum"}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual(expected,
                             ansible_deploy._prepare_variables(task))

    def test__prepare_variables_configdrive_url(self):
        i_info = self.node.instance_info
        i_info['configdrive'] = 'http://configdrive_url'
        self.node.instance_info = i_info
        self.node.save()
        expected = {"image": {"url": "http://image",
                              "validate_certs": "yes",
                              "source": "fake-image",
                              "disk_format": "qcow2",
                              "checksum": "md5:checksum"},
                    'configdrive': {'type': 'url',
                                    'location': 'http://configdrive_url'}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertEqual(expected,
                             ansible_deploy._prepare_variables(task))

    def test__prepare_variables_configdrive_file(self):
        i_info = self.node.instance_info
        i_info['configdrive'] = 'fake-content'
        self.node.instance_info = i_info
        self.node.save()
        configdrive_path = ('%(tempdir)s/%(node)s.cndrive' %
                            {'tempdir': ansible_deploy.CONF.tempdir,
                             'node': self.node.uuid})
        expected = {"image": {"url": "http://image",
                              "validate_certs": "yes",
                              "source": "fake-image",
                              "disk_format": "qcow2",
                              "checksum": "md5:checksum"},
                    'configdrive': {'type': 'file',
                                    'location': configdrive_path}}
        with mock.patch.object(ansible_deploy, 'open', mock.mock_open(),
                               create=True) as open_mock:
            with task_manager.acquire(self.context, self.node.uuid) as task:
                self.assertEqual(expected,
                                 ansible_deploy._prepare_variables(task))
            open_mock.assert_has_calls((
                mock.call(configdrive_path, 'w'),
                mock.call().__enter__(),
                mock.call().write('fake-content'),
                mock.call().__exit__(None, None, None)))

    @mock.patch.object(utils, 'build_configdrive', autospec=True)
    def test__prepare_variables_configdrive_json(self, mock_build_configdrive):
        i_info = self.node.instance_info
        i_info['configdrive'] = {'meta_data': {}}
        self.node.instance_info = i_info
        self.node.save()
        mock_build_configdrive.return_value = 'fake-content'
        configdrive_path = ('%(tempdir)s/%(node)s.cndrive' %
                            {'tempdir': ansible_deploy.CONF.tempdir,
                             'node': self.node.uuid})
        expected = {"image": {"url": "http://image",
                              "validate_certs": "yes",
                              "source": "fake-image",
                              "disk_format": "qcow2",
                              "checksum": "md5:checksum"},
                    'configdrive': {'type': 'file',
                                    'location': configdrive_path}}
        with mock.patch.object(ansible_deploy, 'open', mock.mock_open(),
                               create=True) as open_mock:
            with task_manager.acquire(self.context, self.node.uuid) as task:
                self.assertEqual(expected,
                                 ansible_deploy._prepare_variables(task))
                mock_build_configdrive.assert_called_once_with(
                    task.node, {'meta_data': {}})
            open_mock.assert_has_calls((
                mock.call(configdrive_path, 'w'),
                mock.call().__enter__(),
                mock.call().write('fake-content'),
                mock.call().__exit__(None, None, None)))

    def test__validate_clean_steps(self):
        steps = [{"interface": "deploy",
                  "name": "foo",
                  "args": {"spam": {"required": True, "value": "ham"}}},
                 {"name": "bar",
                  "interface": "deploy"}]
        self.assertIsNone(ansible_deploy._validate_clean_steps(
            steps, self.node.uuid))

    def test__validate_clean_steps_missing(self):
        steps = [{"name": "foo",
                  "interface": "deploy",
                  "args": {"spam": {"value": "ham"},
                           "ham": {"required": True}}},
                 {"name": "bar"},
                 {"interface": "deploy"}]
        exc = self.assertRaises(exception.NodeCleaningFailure,
                                ansible_deploy._validate_clean_steps,
                                steps, self.node.uuid)
        self.assertIn("name foo, field ham.value", str(exc))
        self.assertIn("name bar, field interface", str(exc))
        self.assertIn("name undefined, field name", str(exc))

    def test__validate_clean_steps_names_not_unique(self):
        steps = [{"name": "foo",
                  "interface": "deploy"},
                 {"name": "foo",
                  "interface": "deploy"}]
        exc = self.assertRaises(exception.NodeCleaningFailure,
                                ansible_deploy._validate_clean_steps,
                                steps, self.node.uuid)
        self.assertIn("unique names", str(exc))

    @mock.patch.object(ansible_deploy.yaml, 'safe_load', autospec=True)
    def test__get_clean_steps(self, load_mock):
        steps = [{"interface": "deploy",
                  "name": "foo",
                  "args": {"spam": {"required": True, "value": "ham"}}},
                 {"name": "bar",
                  "interface": "deploy",
                  "priority": 100}]
        load_mock.return_value = steps
        expected = [{"interface": "deploy",
                     "step": "foo",
                     "priority": 10,
                     "abortable": False,
                     "argsinfo": {"spam": {"required": True}},
                     "args": {"spam": "ham"}},
                    {"interface": "deploy",
                     "step": "bar",
                     "priority": 100,
                     "abortable": False,
                     "argsinfo": {},
                     "args": {}}]
        d_info = self.node.driver_info
        d_info['ansible_clean_steps_config'] = 'custom_clean'
        self.node.driver_info = d_info
        self.node.save()
        self.config(group='ansible', playbooks_path='/path/to/playbooks')

        with mock.patch.object(ansible_deploy, 'open', mock.mock_open(),
                               create=True) as open_mock:
            self.assertEqual(
                expected,
                ansible_deploy._get_clean_steps(
                    self.node, interface="deploy",
                    override_priorities={"foo": 10}))
            open_mock.assert_has_calls((
                mock.call('/path/to/playbooks/custom_clean'),))
            load_mock.assert_called_once_with(
                open_mock().__enter__.return_value)


class TestAnsibleDeploy(AnsibleDeployTestCaseBase):
    def setUp(self):
        super(TestAnsibleDeploy, self).setUp()
        self.driver = ansible_deploy.AnsibleDeploy()

    def test_get_properties(self):
        self.assertEqual(
            set(list(ansible_deploy.COMMON_PROPERTIES)
                + ['agent_verify_ca', 'deploy_forces_oob_reboot']),
            set(self.driver.get_properties()))

    @mock.patch.object(deploy_utils, 'check_for_missing_params',
                       autospec=True)
    @mock.patch.object(pxe.PXEBoot, 'validate', autospec=True)
    def test_validate(self, pxe_boot_validate_mock, check_params_mock):
        with task_manager.acquire(
                self.context, self.node['uuid'], shared=False) as task:
            self.driver.validate(task)
            pxe_boot_validate_mock.assert_called_once_with(
                task.driver.boot, task)
            check_params_mock.assert_called_once_with(
                {'instance_info.image_source': INSTANCE_INFO['image_source']},
                mock.ANY)

    @mock.patch.object(ansible_deploy, '_calculate_memory_req', autospec=True,
                       return_value=2000)
    @mock.patch.object(utils, 'node_power_action', autospec=True)
    def test_deploy(self, power_mock, mem_req_mock):
        with task_manager.acquire(
                self.context, self.node['uuid'], shared=False) as task:
            driver_return = self.driver.deploy(task)
            self.assertEqual(driver_return, states.DEPLOYWAIT)
            power_mock.assert_called_once_with(task, states.REBOOT)
            mem_req_mock.assert_called_once_with(task)
            i_info = task.node.instance_info
            self.assertEqual(i_info['image_mem_req'], 2000)

    @mock.patch.object(utils, 'node_power_action', autospec=True)
    def test_tear_down(self, power_mock):
        with task_manager.acquire(
                self.context, self.node['uuid'], shared=False) as task:
            driver_return = self.driver.tear_down(task)
            power_mock.assert_called_once_with(task, states.POWER_OFF)
            self.assertEqual(driver_return, states.DELETED)

    @mock.patch('ironic.conductor.utils.node_power_action', autospec=True)
    @mock.patch('ironic.drivers.modules.deploy_utils.build_agent_options',
                return_value={'op1': 'test1'}, autospec=True)
    @mock.patch('ironic.drivers.modules.deploy_utils.'
                'build_instance_info_for_deploy',
                return_value={'test': 'test'}, autospec=True)
    @mock.patch.object(pxe.PXEBoot, 'prepare_ramdisk', autospec=True)
    def test_prepare(self, pxe_prepare_ramdisk_mock,
                     build_instance_info_mock, build_options_mock,
                     power_action_mock):
        with task_manager.acquire(
                self.context, self.node['uuid'], shared=False) as task:
            task.node.provision_state = states.DEPLOYING

            with mock.patch.object(task.driver.network,
                                   'add_provisioning_network',
                                   autospec=True) as net_mock:
                self.driver.prepare(task)

                net_mock.assert_called_once_with(task)
            power_action_mock.assert_called_once_with(task,
                                                      states.POWER_OFF)
            build_instance_info_mock.assert_called_once_with(task)
            build_options_mock.assert_called_once_with(task.node)
            pxe_prepare_ramdisk_mock.assert_called_once_with(
                task.driver.boot, task, {'op1': 'test1'})

        self.node.refresh()
        self.assertEqual('test', self.node.instance_info['test'])

    @mock.patch.object(ansible_deploy, '_get_configdrive_path',
                       return_value='/path/test', autospec=True)
    @mock.patch.object(irlib_utils, 'unlink_without_raise', autospec=True)
    @mock.patch.object(pxe.PXEBoot, 'clean_up_ramdisk', autospec=True)
    def test_clean_up(self, pxe_clean_up_mock, unlink_mock,
                      get_cfdrive_path_mock):
        with task_manager.acquire(
                self.context, self.node['uuid'], shared=False) as task:
            self.driver.clean_up(task)
            pxe_clean_up_mock.assert_called_once_with(task.driver.boot, task)
            get_cfdrive_path_mock.assert_called_once_with(self.node['uuid'])
            unlink_mock.assert_called_once_with('/path/test')

    @mock.patch.object(ansible_deploy, '_get_clean_steps', autospec=True)
    def test_get_clean_steps(self, get_clean_steps_mock):
        mock_steps = [{'priority': 10, 'interface': 'deploy',
                       'step': 'erase_devices'},
                      {'priority': 99, 'interface': 'deploy',
                       'step': 'erase_devices_metadata'},
                      ]
        get_clean_steps_mock.return_value = mock_steps
        with task_manager.acquire(self.context, self.node.uuid) as task:
            steps = self.driver.get_clean_steps(task)
            get_clean_steps_mock.assert_called_once_with(
                task.node, interface='deploy',
                override_priorities={
                    'erase_devices': None,
                    'erase_devices_metadata': None})
        self.assertEqual(mock_steps, steps)

    @mock.patch.object(ansible_deploy, '_get_clean_steps', autospec=True)
    def test_get_clean_steps_priority(self, mock_get_clean_steps):
        self.config(erase_devices_priority=9, group='deploy')
        self.config(erase_devices_metadata_priority=98, group='deploy')
        mock_steps = [{'priority': 9, 'interface': 'deploy',
                       'step': 'erase_devices'},
                      {'priority': 98, 'interface': 'deploy',
                       'step': 'erase_devices_metadata'},
                      ]
        mock_get_clean_steps.return_value = mock_steps

        with task_manager.acquire(self.context, self.node.uuid) as task:
            steps = self.driver.get_clean_steps(task)
            mock_get_clean_steps.assert_called_once_with(
                task.node, interface='deploy',
                override_priorities={'erase_devices': 9,
                                     'erase_devices_metadata': 98})
        self.assertEqual(mock_steps, steps)

    @mock.patch.object(ansible_deploy, '_run_playbook', autospec=True)
    @mock.patch.object(ansible_deploy, '_prepare_extra_vars', autospec=True)
    @mock.patch.object(ansible_deploy, '_parse_ansible_driver_info',
                       return_value=('test_pl', 'test_u', 'test_k'),
                       autospec=True)
    def test_execute_clean_step(self, parse_driver_info_mock,
                                prepare_extra_mock, run_playbook_mock):

        step = {'priority': 10, 'interface': 'deploy',
                'step': 'erase_devices', 'args': {'tags': ['clean']}}
        ironic_nodes = {
            'ironic_nodes': [(self.node['uuid'], '127.0.0.1', 'test_u', {})]}
        prepare_extra_mock.return_value = ironic_nodes
        di_info = self.node.driver_internal_info
        di_info['agent_url'] = 'http://127.0.0.1'
        self.node.driver_internal_info = di_info
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.execute_clean_step(task, step)

            parse_driver_info_mock.assert_called_once_with(
                task.node, action='clean')
            prepare_extra_mock.assert_called_once_with(
                ironic_nodes['ironic_nodes'])
            run_playbook_mock.assert_called_once_with(
                task.node, 'test_pl', ironic_nodes, 'test_k', tags=['clean'])

    @mock.patch.object(ansible_deploy, '_parse_ansible_driver_info',
                       return_value=('test_pl', 'test_u', 'test_k'),
                       autospec=True)
    @mock.patch.object(ansible_deploy, '_run_playbook', autospec=True)
    @mock.patch.object(ansible_deploy, 'LOG', autospec=True)
    def test_execute_clean_step_no_success_log(
            self, log_mock, run_mock, parse_driver_info_mock):

        run_mock.side_effect = exception.InstanceDeployFailure('Boom')
        step = {'priority': 10, 'interface': 'deploy',
                'step': 'erase_devices', 'args': {'tags': ['clean']}}
        di_info = self.node.driver_internal_info
        di_info['agent_url'] = 'http://127.0.0.1'
        self.node.driver_internal_info = di_info
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.InstanceDeployFailure,
                              self.driver.execute_clean_step,
                              task, step)
            self.assertFalse(log_mock.info.called)

    @mock.patch.object(ansible_deploy, '_run_playbook', autospec=True)
    @mock.patch.object(steps, 'set_node_cleaning_steps', autospec=True)
    @mock.patch.object(utils, 'node_power_action', autospec=True)
    @mock.patch('ironic.drivers.modules.deploy_utils.build_agent_options',
                return_value={'op1': 'test1'}, autospec=True)
    @mock.patch.object(pxe.PXEBoot, 'prepare_ramdisk', autospec=True)
    def test_prepare_cleaning(
            self, prepare_ramdisk_mock, buid_options_mock, power_action_mock,
            set_node_cleaning_steps, run_playbook_mock):
        step = {'priority': 10, 'interface': 'deploy',
                'step': 'erase_devices', 'tags': ['clean']}
        driver_internal_info = dict(DRIVER_INTERNAL_INFO)
        driver_internal_info['clean_steps'] = [step]
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.network.add_cleaning_network = mock.Mock()

            state = self.driver.prepare_cleaning(task)

            set_node_cleaning_steps.assert_called_once_with(task)
            task.driver.network.add_cleaning_network.assert_called_once_with(
                task)
            buid_options_mock.assert_called_once_with(task.node)
            prepare_ramdisk_mock.assert_called_once_with(
                task.driver.boot, task, {'op1': 'test1'})
            power_action_mock.assert_called_once_with(task, states.REBOOT)
            self.assertFalse(run_playbook_mock.called)
            self.assertEqual(states.CLEANWAIT, state)

    @mock.patch.object(steps, 'set_node_cleaning_steps', autospec=True)
    def test_prepare_cleaning_callback_no_steps(self,
                                                set_node_cleaning_steps):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.network.add_cleaning_network = mock.Mock()

            self.driver.prepare_cleaning(task)

            set_node_cleaning_steps.assert_called_once_with(task)
            self.assertFalse(task.driver.network.add_cleaning_network.called)

    @mock.patch.object(utils, 'node_power_action', autospec=True)
    @mock.patch.object(pxe.PXEBoot, 'clean_up_ramdisk', autospec=True)
    def test_tear_down_cleaning(self, clean_ramdisk_mock, power_action_mock):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.network.remove_cleaning_network = mock.Mock()

            self.driver.tear_down_cleaning(task)

            power_action_mock.assert_called_once_with(task, states.POWER_OFF)
            clean_ramdisk_mock.assert_called_once_with(task.driver.boot, task)
            (task.driver.network.remove_cleaning_network
                .assert_called_once_with(task))

    @mock.patch.object(ansible_deploy, '_run_playbook', autospec=True)
    @mock.patch.object(ansible_deploy, '_prepare_extra_vars', autospec=True)
    @mock.patch.object(ansible_deploy, '_parse_ansible_driver_info',
                       return_value=('test_pl', 'test_u', 'test_k'),
                       autospec=True)
    @mock.patch.object(ansible_deploy, '_parse_partitioning_info',
                       autospec=True)
    @mock.patch.object(ansible_deploy, '_prepare_variables', autospec=True)
    def test__ansible_deploy(self, prepare_vars_mock, parse_part_info_mock,
                             parse_dr_info_mock, prepare_extra_mock,
                             run_playbook_mock):
        ironic_nodes = {
            'ironic_nodes': [(self.node['uuid'], '127.0.0.1', 'test_u')]}
        prepare_extra_mock.return_value = ironic_nodes
        _vars = {
            'url': 'image_url',
            'checksum': 'aa'}
        prepare_vars_mock.return_value = _vars

        driver_internal_info = dict(DRIVER_INTERNAL_INFO)
        driver_internal_info['is_whole_disk_image'] = False
        self.node.driver_internal_info = driver_internal_info
        self.node.extra = {'ham': 'spam'}
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver._ansible_deploy(task, '127.0.0.1')

            prepare_vars_mock.assert_called_once_with(task)
            parse_part_info_mock.assert_called_once_with(task.node)
            parse_dr_info_mock.assert_called_once_with(task.node)
            prepare_extra_mock.assert_called_once_with(
                [(self.node['uuid'], '127.0.0.1', 'test_u', {'ham': 'spam'})],
                variables=_vars)
            run_playbook_mock.assert_called_once_with(
                task.node, 'test_pl', ironic_nodes, 'test_k')

    @mock.patch.object(ansible_deploy, '_run_playbook', autospec=True)
    @mock.patch.object(ansible_deploy, '_prepare_extra_vars', autospec=True)
    @mock.patch.object(ansible_deploy, '_parse_ansible_driver_info',
                       return_value=('test_pl', 'test_u', 'test_k'),
                       autospec=True)
    @mock.patch.object(ansible_deploy, '_parse_partitioning_info',
                       autospec=True)
    @mock.patch.object(ansible_deploy, '_prepare_variables', autospec=True)
    def test__ansible_deploy_iwdi(self, prepare_vars_mock,
                                  parse_part_info_mock, parse_dr_info_mock,
                                  prepare_extra_mock, run_playbook_mock):
        ironic_nodes = {
            'ironic_nodes': [(self.node['uuid'], '127.0.0.1', 'test_u')]}
        prepare_extra_mock.return_value = ironic_nodes
        _vars = {
            'url': 'image_url',
            'checksum': 'aa'}
        prepare_vars_mock.return_value = _vars
        driver_internal_info = self.node.driver_internal_info
        driver_internal_info['is_whole_disk_image'] = True
        instance_info = self.node.instance_info
        del instance_info['root_mb']
        self.node.driver_internal_info = driver_internal_info
        self.node.instance_info = instance_info
        self.node.extra = {'ham': 'spam'}
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver._ansible_deploy(task, '127.0.0.1')

            prepare_vars_mock.assert_called_once_with(task)
            self.assertFalse(parse_part_info_mock.called)
            parse_dr_info_mock.assert_called_once_with(task.node)
            prepare_extra_mock.assert_called_once_with(
                [(self.node['uuid'], '127.0.0.1', 'test_u', {'ham': 'spam'})],
                variables=_vars)
            run_playbook_mock.assert_called_once_with(
                task.node, 'test_pl', ironic_nodes, 'test_k')

    @mock.patch.object(fake.FakePower, 'get_power_state',
                       return_value=states.POWER_OFF, autospec=True)
    @mock.patch.object(utils, 'node_power_action', autospec=True)
    def test_tear_down_agent_force_reboot(
            self, power_action_mock, get_pow_state_mock):
        d_info = self.node.driver_info
        d_info['deploy_forces_oob_reboot'] = True
        self.node.driver_info = d_info
        self.node.save()
        self.config(group='ansible',
                    post_deploy_get_power_state_retry_interval=0)
        self.node.provision_state = states.DEPLOYING
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.tear_down_agent(task)
            power_action_mock.assert_called_once_with(task, states.POWER_OFF)
        get_pow_state_mock.assert_not_called()

    @mock.patch.object(ansible_deploy, '_run_playbook', autospec=True)
    @mock.patch.object(utils, 'node_power_action', autospec=True)
    def test_tear_down_agent_soft_poweroff_retry(
            self, power_action_mock, run_playbook_mock):
        self.config(group='ansible',
                    post_deploy_get_power_state_retry_interval=0)
        self.config(group='ansible',
                    post_deploy_get_power_state_retries=1)
        self.node.provision_state = states.DEPLOYING
        di_info = self.node.driver_internal_info
        di_info['agent_url'] = 'http://127.0.0.1'
        self.node.driver_internal_info = di_info
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            with mock.patch.object(task.driver.power,
                                   'get_power_state',
                                   return_value=states.POWER_ON,
                                   autospec=True) as p_mock:
                self.driver.tear_down_agent(task)
                p_mock.assert_called_with(task)
                self.assertEqual(2, len(p_mock.mock_calls))
            power_action_mock.assert_called_once_with(task, states.POWER_OFF)
            run_playbook_mock.assert_called_once_with(
                task.node, 'shutdown.yaml', mock.ANY, mock.ANY)

    @mock.patch.object(utils, 'node_set_boot_device', autospec=True)
    @mock.patch.object(ansible_deploy, '_get_node_ip', autospec=True,
                       return_value='1.2.3.4')
    def test_write_image(self, getip_mock, bootdev_mock):
        self.node.provision_state = states.DEPLOYING
        self.node.target_provision_state = states.ACTIVE
        self.node.save()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            with mock.patch.multiple(self.driver, autospec=True,
                                     _ansible_deploy=mock.DEFAULT,
                                     reboot_to_instance=mock.DEFAULT):
                result = self.driver.write_image(task)
                self.assertIsNone(result)
                getip_mock.assert_called_once_with(task)
                self.driver._ansible_deploy.assert_called_once_with(
                    task, '1.2.3.4')
                bootdev_mock.assert_called_once_with(task, 'disk',
                                                     persistent=True)
            self.assertEqual(states.ACTIVE, task.node.target_provision_state)
            self.assertEqual(states.DEPLOYING, task.node.provision_state)

    @mock.patch.object(flat_network.FlatNetwork, 'add_provisioning_network',
                       autospec=True)
    @mock.patch.object(utils, 'restore_power_state_if_needed', autospec=True)
    @mock.patch.object(utils, 'power_on_node_if_needed', autospec=True)
    @mock.patch.object(utils, 'node_power_action', autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(deploy_utils, 'build_instance_info_for_deploy',
                       autospec=True)
    @mock.patch.object(pxe.PXEBoot, 'prepare_ramdisk', autospec=True)
    def test_prepare_with_smartnic_port(
            self, pxe_prepare_ramdisk_mock,
            build_instance_info_mock, build_options_mock,
            power_action_mock, power_on_node_if_needed_mock,
            restore_power_state_mock, net_mock):
        with task_manager.acquire(
                self.context, self.node['uuid'], shared=False) as task:
            task.node.provision_state = states.DEPLOYING
            build_instance_info_mock.return_value = {'test': 'test'}
            build_options_mock.return_value = {'op1': 'test1'}
            power_on_node_if_needed_mock.return_value = states.POWER_OFF
            self.driver.prepare(task)
            power_action_mock.assert_called_once_with(
                task, states.POWER_OFF)
            build_instance_info_mock.assert_called_once_with(task)
            build_options_mock.assert_called_once_with(task.node)
            pxe_prepare_ramdisk_mock.assert_called_once_with(
                task.driver.boot, task, {'op1': 'test1'})
            power_on_node_if_needed_mock.assert_called_once_with(task)
            restore_power_state_mock.assert_called_once_with(
                task, states.POWER_OFF)

        self.node.refresh()
        self.assertEqual('test', self.node.instance_info['test'])

    @mock.patch.object(utils, 'restore_power_state_if_needed', autospec=True)
    @mock.patch.object(utils, 'power_on_node_if_needed', autospec=True)
    @mock.patch.object(ansible_deploy, '_run_playbook', autospec=True)
    @mock.patch.object(steps, 'set_node_cleaning_steps', autospec=True)
    @mock.patch.object(utils, 'node_power_action', autospec=True)
    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(pxe.PXEBoot, 'prepare_ramdisk', autospec=True)
    def test_prepare_cleaning_with_smartnic_port(
            self, prepare_ramdisk_mock, build_options_mock, power_action_mock,
            set_node_cleaning_steps, run_playbook_mock,
            power_on_node_if_needed_mock, restore_power_state_mock):
        step = {'priority': 10, 'interface': 'deploy',
                'step': 'erase_devices', 'tags': ['clean']}
        driver_internal_info = dict(DRIVER_INTERNAL_INFO)
        driver_internal_info['clean_steps'] = [step]
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.network.add_cleaning_network = mock.Mock()
            build_options_mock.return_value = {'op1': 'test1'}
            power_on_node_if_needed_mock.return_value = states.POWER_OFF
            state = self.driver.prepare_cleaning(task)
            set_node_cleaning_steps.assert_called_once_with(task)
            task.driver.network.add_cleaning_network.assert_called_once_with(
                task)
            build_options_mock.assert_called_once_with(task.node)
            prepare_ramdisk_mock.assert_called_once_with(
                task.driver.boot, task, {'op1': 'test1'})
            power_action_mock.assert_called_once_with(task, states.REBOOT)
            self.assertFalse(run_playbook_mock.called)
            self.assertEqual(states.CLEANWAIT, state)
            power_on_node_if_needed_mock.assert_called_once_with(task)
            restore_power_state_mock.assert_called_once_with(
                task, states.POWER_OFF)

    @mock.patch.object(utils, 'restore_power_state_if_needed', autospec=True)
    @mock.patch.object(utils, 'power_on_node_if_needed', autospec=True)
    @mock.patch.object(utils, 'node_power_action', autospec=True)
    @mock.patch.object(pxe.PXEBoot, 'clean_up_ramdisk', autospec=True)
    def test_tear_down_cleaning_with_smartnic_port(
            self, clean_ramdisk_mock, power_action_mock,
            power_on_node_if_needed_mock, restore_power_state_mock):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.network.remove_cleaning_network = mock.Mock()
            power_on_node_if_needed_mock.return_value = states.POWER_OFF
            self.driver.tear_down_cleaning(task)
            power_action_mock.assert_called_once_with(task, states.POWER_OFF)
            power_action_mock.assert_called_once_with(task, states.POWER_OFF)
            clean_ramdisk_mock.assert_called_once_with(task.driver.boot, task)
            (task.driver.network.remove_cleaning_network
                .assert_called_once_with(task))
            power_on_node_if_needed_mock.assert_called_once_with(task)
            restore_power_state_mock.assert_called_once_with(
                task, states.POWER_OFF)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import pxe_utils as common_pxe_utils
from ironic.drivers import base

CONF = cfg.CONF

LOG = log.getLogger(__name__)


class ExternalStorage(base.StorageInterface):
    """Externally driven Storage Interface."""

    def validate(self, task):
        def _fail_validation(task, reason,
                             exception=exception.InvalidParameterValue):
            msg = (_("Failed to validate external storage interface for node "
                     "%(node)s. %(reason)s") %
                   {'node': task.node.uuid, 'reason': reason})
            LOG.error(msg)
            raise exception(msg)

        if (not self.should_write_image(task)
                and not common_pxe_utils.is_ipxe_enabled(task)):
            msg = _("The [pxe]/ipxe_enabled option must "
                    "be set to True to support network "
                    "booting to an iSCSI volume or the boot "
                    "interface must be set to ``ipxe``.")
            _fail_validation(task, msg)

    def get_properties(self):
        return {}

    def attach_volumes(self, task):
        pass

    def detach_volumes(self, task):
        pass

    def should_write_image(self, task):
        """Determines if deploy should perform the image write-out.

        This enables the user to define a volume and Ironic understand
        that the image may already exist and we may be booting to that volume.

        :param task: The task object.
        :returns: True if the deployment write-out process should be
                  executed.
        """
        instance_info = task.node.instance_info
        if 'image_source' not in instance_info:
            for volume in task.volume_targets:
                if volume['boot_index'] == 0:
                    return False
        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
The agent power interface.
"""

import time

from oslo_config import cfg
from oslo_log import log
import tenacity

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.common import utils
from ironic.conductor import utils as cond_utils
from ironic.drivers import base
from ironic.drivers.modules import agent_client


CONF = cfg.CONF

LOG = log.getLogger(__name__)

_POWER_WAIT = 30


class AgentPower(base.PowerInterface):
    """Power interface using the running agent for power actions."""

    def __init__(self):
        super(AgentPower, self).__init__()
        self._client = agent_client.AgentClient()

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return {}

    def validate(self, task):
        """Validate the driver-specific Node deployment info.

        :param task: A TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue on malformed parameter(s)
        """
        # NOTE(dtantsur): the fast_track option is mutable, so we have to check
        # it again on validation.
        if not utils.fast_track_enabled(task.node):
            raise exception.InvalidParameterValue(
                _('Fast track mode must be enabled to use the agent '
                  'power interface'))
        # TODO(dtantsur): support ACTIVE nodes
        if not cond_utils.agent_is_alive(task.node):
            raise exception.InvalidParameterValue(
                _('Agent seems offline for node %s, the agent power interface '
                  'cannot be used') % task.node.uuid)

    def supports_power_sync(self, task):
        """Check if power sync is supported for the given node.

        Not supported for the agent power since it is not possible to power
        on/off nodes.

        :param task: A TaskManager instance containing the node to act on
            with a **shared** lock.
        :returns: boolean, whether power sync is supported.
        """
        return False

    def get_supported_power_states(self, task):
        """Get a list of the supported power states.

        Only contains REBOOT.

        :param task: A TaskManager instance containing the node to act on.
        :returns: A list with the supported power states defined
                  in :mod:`ironic.common.states`.
        """
        return [states.REBOOT, states.SOFT_REBOOT]

    def get_power_state(self, task):
        """Return the power state of the task's node.

        Essentially, the only known state is POWER ON, everything else is
        an error (or more precisely ``None``).

        :param task: A TaskManager instance containing the node to act on.
        :returns: A power state. One of :mod:`ironic.common.states`.
        """
        # TODO(dtantsur): support ACTIVE nodes
        if cond_utils.agent_is_alive(task.node):
            return states.POWER_ON
        else:
            LOG.error('Node %s is not fast-track-able, cannot determine '
                      'its power state via the "agent" power interface',
                      task.node.uuid)
            return None

    def set_power_state(self, task, power_state, timeout=None):
        """Set the power state of the task's node.

        :param task: A TaskManager instance containing the node to act on.
        :param power_state: Power state from :mod:`ironic.common.states`.
            Only REBOOT and SOFT_REBOOT are supported and are synonymous.
        :param timeout: timeout (in seconds) positive integer (> 0) for any
          power state. ``None`` indicates to use default timeout.
        :raises: PowerStateFailure on non-supported power state.
        """
        if power_state in (states.REBOOT, states.SOFT_REBOOT):
            return self.reboot(task)
        else:
            LOG.error('Power state %(state)s is not implemented for node '
                      '%(node)s using the "agent" power interface',
                      {'node': task.node.uuid, 'state': power_state})
            raise exception.PowerStateFailure(pstate=power_state)

    def reboot(self, task, timeout=None):
        """Perform a reboot of the task's node.

        Only soft reboot is implemented.

        :param task: A TaskManager instance containing the node to act on.
        :param timeout: timeout (in seconds) positive integer (> 0) for any
            power state. ``None`` indicates to use default timeout.
        """
        node = task.node

        self._client.reboot(node)

        # NOTE(dtantsur): wipe the agent token, otherwise the rebooted agent
        # won't be able to heartbeat. This is mostly a precaution since the
        # calling code in conductor is expected to handle it.
        if not node.driver_internal_info.get(
                'agent_secret_token_pregenerated'):
            node.del_driver_internal_info('agent_secret_token')
        # NOTE(dtantsur): the URL may change on reboot, wipe it as well (but
        # only after we call reboot).
        node.del_driver_internal_info('agent_url')
        node.save()

        LOG.debug('Requested reboot of node %(node)s via the agent, waiting '
                  '%(wait)d seconds for the node to power down',
                  {'node': task.node.uuid, 'wait': _POWER_WAIT})
        time.sleep(_POWER_WAIT)

        if (node.provision_state in (states.DEPLOYING, states.CLEANING)
                and (node.driver_internal_info.get('deployment_reboot')
                     or node.driver_internal_info.get('cleaning_reboot'))):
            # NOTE(dtantsur): we need to downgrade the lock otherwise
            # heartbeats won't be processed. It should not have side effects
            # for nodes in DEPLOYING/CLEANING.
            task.downgrade_lock()

            try:
                self._wait_for_reboot(task, timeout)
            finally:
                # The caller probably expects a lock, so re-acquire it
                task.upgrade_lock()

    def _wait_for_reboot(self, task, timeout):
        wait = CONF.agent.post_deploy_get_power_state_retry_interval
        if not timeout:
            timeout = CONF.agent.post_deploy_get_power_state_retries * wait

        @tenacity.retry(
            stop=tenacity.stop_after_delay(timeout),
            retry=(tenacity.retry_if_result(lambda result: not result)
                   | tenacity.retry_if_exception_type(
                exception.AgentConnectionFailed)),
            wait=tenacity.wait_fixed(wait),
            reraise=True)
        def _wait_until_rebooted(task):
            try:
                status = self._client.get_commands_status(
                    task.node, retry_connection=False, expect_errors=True)
            except exception.AgentConnectionFailed:
                LOG.debug('Still waiting for the agent to come back on the '
                          'node %s', task.node.uuid)
                raise

            if any(cmd['command_name'] == agent_client.REBOOT_COMMAND
                   for cmd in status):
                LOG.debug('Still waiting for the agent to power off on the '
                          'node %s', task.node.uuid)
                return False

            return True

        try:
            _wait_until_rebooted(task)
        except exception.AgentConnectionFailed as exc:
            msg = _('Agent failed to come back on %(node)s with the "agent" '
                    'power interface: %(exc)s') % {
                        'node': task.node.uuid, 'exc': exc}
            LOG.error(msg)
            raise exception.PowerStateFailure(msg)
        except Exception as exc:
            LOG.error('Could not reboot node %(node)s with the "agent" power '
                      'interface: %(exc)s',
                      {'node': task.node.uuid, 'exc': exc})
            raise exception.PowerStateFailure(
                _('Unexpected error when rebooting through the agent: %s')
                % exc)



# -*- encoding: utf-8 -*-
#
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Base classes for API tests."""

# NOTE: Ported from ceilometer/tests/api.py (subsequently moved to
#       ceilometer/tests/api/__init__.py). This should be oslo'ified:
#       https://bugs.launchpad.net/ironic/+bug/1255115.

from unittest import mock
from urllib import parse as urlparse

from oslo_config import cfg
import pecan
import pecan.testing

from ironic.api.controllers.v1 import node as api_node
from ironic.tests.unit.db import base as db_base

PATH_PREFIX = '/v1'

cfg.CONF.import_group('keystone_authtoken', 'keystonemiddleware.auth_token')


class BaseApiTest(db_base.DbTestCase):
    """Pecan controller functional testing class.

    Used for functional tests of Pecan controllers where you need to
    test your literal application and its integration with the
    framework.
    """

    SOURCE_DATA = {'test_source': {'somekey': '666'}}

    root_controller = 'ironic.api.controllers.root.RootController'

    def setUp(self):
        super(BaseApiTest, self).setUp()
        cfg.CONF.set_override("auth_version", "v3",
                              group='keystone_authtoken')
        cfg.CONF.set_override("admin_user", "admin",
                              group='keystone_authtoken')
        cfg.CONF.set_override("auth_strategy", "noauth")
        self.app = self._make_app()

        def reset_pecan():
            pecan.set_config({}, overwrite=True)

        self.addCleanup(reset_pecan)

        p = mock.patch('ironic.api.controllers.v1.Controller._check_version',
                       autospec=True)
        self._check_version = p.start()
        self.addCleanup(p.stop)

        api_node._NETWORK_DATA_SCHEMA = None
        api_node._NODE_VALIDATOR = None
        api_node._NODE_PATCH_VALIDATOR = None

    def _make_app(self):
        # Determine where we are so we can set up paths in the config
        root_dir = self.path_get()

        self.app_config = {
            'app': {
                'root': self.root_controller,
                'modules': ['ironic.api'],
                'static_root': '%s/public' % root_dir,
                'debug': True,
                'template_path': '%s/api/templates' % root_dir,
                'acl_public_routes': ['/', '/v1'],
            },
        }
        return pecan.testing.load_test_app(self.app_config)

    def _request_json(self, path, params, expect_errors=False, headers=None,
                      method="post", extra_environ=None, status=None,
                      path_prefix=PATH_PREFIX):
        """Sends simulated HTTP request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param method: Request method type. Appropriate method function call
                       should be used rather than passing attribute in.
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        :param path_prefix: prefix of the url path
        """
        full_path = path_prefix + path
        print('%s: %s %s' % (method.upper(), full_path, params))
        response = getattr(self.app, "%s_json" % method)(
            str(full_path),
            params=params,
            headers=headers,
            status=status,
            extra_environ=extra_environ,
            expect_errors=expect_errors
        )
        print('GOT:%s' % response)
        return response

    def put_json(self, path, params, expect_errors=False, headers=None,
                 extra_environ=None, status=None, path_prefix=PATH_PREFIX):
        """Sends simulated HTTP PUT request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        return self._request_json(path=path, params=params,
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method="put",
                                  path_prefix=path_prefix)

    def post_json(self, path, params, expect_errors=False, headers=None,
                  extra_environ=None, status=None, path_prefix=PATH_PREFIX):
        """Sends simulated HTTP POST request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        return self._request_json(path=path, params=params,
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method="post",
                                  path_prefix=path_prefix)

    def patch_json(self, path, params, expect_errors=False, headers=None,
                   extra_environ=None, status=None, path_prefix=PATH_PREFIX):
        """Sends simulated HTTP PATCH request to Pecan test app.

        :param path: url path of target service
        :param params: content for wsgi.input of request
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        """
        return self._request_json(path=path, params=params,
                                  expect_errors=expect_errors,
                                  headers=headers, extra_environ=extra_environ,
                                  status=status, method="patch",
                                  path_prefix=path_prefix)

    def delete(self, path, expect_errors=False, headers=None,
               extra_environ=None, status=None, path_prefix=PATH_PREFIX):
        """Sends simulated HTTP DELETE request to Pecan test app.

        :param path: url path of target service
        :param expect_errors: Boolean value; whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param status: expected status code of response
        :param path_prefix: prefix of the url path
        """
        full_path = path_prefix + path
        print('DELETE: %s' % (full_path))
        response = self.app.delete(str(full_path),
                                   headers=headers,
                                   status=status,
                                   extra_environ=extra_environ,
                                   expect_errors=expect_errors)
        print('GOT:%s' % response)
        return response

    def get_json(self, path, expect_errors=False, headers=None,
                 extra_environ=None, q=None, path_prefix=PATH_PREFIX,
                 **params):
        """Sends simulated HTTP GET request to Pecan test app.

        :param path: url path of target service
        :param expect_errors: Boolean value;whether an error is expected based
                              on request
        :param headers: a dictionary of headers to send along with the request
        :param extra_environ: a dictionary of environ variables to send along
                              with the request
        :param q: list of queries consisting of: field, value, op, and type
                  keys
        :param path_prefix: prefix of the url path
        :param params: content for wsgi.input of request
        """
        q = q if q is not None else []
        full_path = path_prefix + path
        query_params = {'q.field': [],
                        'q.value': [],
                        'q.op': [],
                        }
        for query in q:
            for name in ['field', 'op', 'value']:
                query_params['q.%s' % name].append(query.get(name, ''))
        all_params = {}
        all_params.update(params)
        if q:
            all_params.update(query_params)
        print('GET: %s %r' % (full_path, all_params))
        response = self.app.get(full_path,
                                params=all_params,
                                headers=headers,
                                extra_environ=extra_environ,
                                expect_errors=expect_errors)
        if not expect_errors:
            response = response.json
        print('GOT:%s' % response)
        return response

    def validate_link(self, link, bookmark=False, headers=None):
        """Checks if the given link can get correct data."""
        # removes the scheme and net location parts of the link
        url_parts = list(urlparse.urlparse(link))
        url_parts[0] = url_parts[1] = ''

        # bookmark link should not have the version in the URL
        if bookmark and url_parts[2].startswith(PATH_PREFIX):
            return False

        full_path = urlparse.urlunparse(url_parts)
        try:
            self.get_json(full_path, path_prefix='', headers=headers)
            return True
        except Exception:
            return False



# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
iRMC Management Driver
"""

from ironic_lib import metrics_utils
from oslo_log import log as logging
from oslo_utils import importutils

from ironic.common import boot_devices
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic import conf
from ironic.drivers import base
from ironic.drivers.modules import boot_mode_utils
from ironic.drivers.modules import ipmitool
from ironic.drivers.modules.irmc import common as irmc_common
from ironic.drivers.modules.redfish import management as redfish_management

irmc = importutils.try_import('scciclient.irmc')

LOG = logging.getLogger(__name__)
CONF = conf.CONF

METRICS = metrics_utils.get_metrics_logger(__name__)

# Boot Option Parameters #5 Data2 defined in
# Set/Get System Boot Options Command, IPMI spec v2.0.
_BOOTPARAM5_DATA2 = {boot_devices.PXE: '0x04',
                     boot_devices.DISK: '0x08',
                     # note (naohirot)
                     # boot_devices.CDROM is tentatively set to '0x20' rather
                     # than '0x14' as a work-around to force iRMC vmedia boot.
                     #   0x14 = Force boot from default CD/DVD
                     #   0x20 = Force boot from remotely connected CD/DVD
                     boot_devices.CDROM: '0x20',
                     boot_devices.BIOS: '0x18',
                     boot_devices.SAFE: '0x0c',
                     }


def _get_sensors_data(task):
    """Get sensors data method.

    It gets sensor data from the task's node via SCCI, and convert the data
    from XML to the dict format.

    :param task: A TaskManager instance.
    :raises: FailedToGetSensorData when getting the sensor data fails.
    :returns: Returns a consistent formatted dict of sensor data grouped
              by sensor type, which can be processed by Ceilometer.
    """

    try:
        report = irmc_common.get_irmc_report(task.node)
        sensor = irmc.scci.get_sensor_data(report)

    except (exception.InvalidParameterValue,
            exception.MissingParameterValue,
            irmc.scci.SCCIInvalidInputError,
            irmc.scci.SCCIClientError) as e:
        LOG.error("SCCI get sensor data failed for node %(node_id)s "
                  "with the following error: %(error)s",
                  {'node_id': task.node.uuid, 'error': e})
        raise exception.FailedToGetSensorData(
            node=task.node.uuid, error=e)

    sensors_data = {}
    for sdr in sensor:
        sensor_type_name = sdr.find('./Data/Decoded/Sensor/TypeName')
        sensor_type_number = sdr.find('./Data/Decoded/Sensor/Type')
        entity_name = sdr.find('./Data/Decoded/Entity/Name')
        entity_id = sdr.find('./Data/Decoded/Entity/ID')

        if None in (sensor_type_name, sensor_type_number,
                    entity_name, entity_id):
            continue

        sensor_type = ('%s (%s)' %
                       (sensor_type_name.text, sensor_type_number.text))
        sensor_id = ('%s (%s)' %
                     (entity_name.text, entity_id.text))
        reading_value = sdr.find(
            './Data/Decoded/Sensor/Thresholds/*/Normalized')
        reading_value_text = "None" if (
            reading_value is None) else str(reading_value.text)
        reading_units = sdr.find('./Data/Decoded/Sensor/BaseUnitName')
        reading_units_text = "None" if (
            reading_units is None) else str(reading_units.text)
        sensor_reading = '%s %s' % (reading_value_text, reading_units_text)

        sensors_data.setdefault(sensor_type, {})[sensor_id] = {
            'Sensor Reading': sensor_reading,
            'Sensor ID': sensor_id,
            'Units': reading_units_text,
        }

    return sensors_data


def backup_bios_config(task):
    """Backup BIOS config from a node.

    :param task: a TaskManager instance containing the node to act on.
    :raises: IRMCOperationError on failure.
    """
    node_uuid = task.node.uuid

    # Skip this operation if the clean step 'restore' is disabled
    if CONF.irmc.clean_priority_restore_irmc_bios_config == 0:
        LOG.debug('Skipped the operation backup_BIOS_config for node %s '
                  'as the clean step restore_BIOS_config is disabled.',
                  node_uuid)
        return

    irmc_info = irmc_common.parse_driver_info(task.node)

    try:
        # Backup bios config
        result = irmc.elcm.backup_bios_config(irmc_info)
    except irmc.scci.SCCIError as e:
        LOG.error('Failed to backup BIOS config for node %(node)s. '
                  'Error: %(error)s', {'node': node_uuid, 'error': e})
        raise exception.IRMCOperationError(operation='backup BIOS config',
                                           error=e)

    # Save bios config into the driver_internal_info
    task.node.set_driver_internal_info('irmc_bios_config',
                                       result['bios_config'])
    task.node.save()

    LOG.info('BIOS config is backed up successfully for node %s',
             node_uuid)

    # NOTE(tiendc): When the backup operation done, server is automatically
    # shutdown. However, this function is called right before the method
    # task.driver.deploy() that will trigger a reboot. So, we don't need
    # to power on the server at this point.


def _restore_bios_config(task):
    """Restore BIOS config to a node.

    :param task: a TaskManager instance containing the node to act on.
    :raises: IRMCOperationError if the operation fails.
    """
    node_uuid = task.node.uuid

    # Get bios config stored in the node object
    bios_config = task.node.driver_internal_info.get('irmc_bios_config')
    if not bios_config:
        LOG.info('Skipped operation "restore BIOS config" on node %s '
                 'as the backup data not found.', node_uuid)
        return

    def _remove_bios_config(task, reboot_flag=False):
        """Remove backup bios config from the node."""
        task.node.del_driver_internal_info('irmc_bios_config')
        # NOTE(tiendc): If reboot flag is raised, then the BM will
        # reboot and cause a bug if the next clean step is in-band.
        # See https://storyboard.openstack.org/#!/story/2002731
        if reboot_flag:
            task.node.set_driver_internal_info('cleaning_reboot', True)
        task.node.save()

    irmc_info = irmc_common.parse_driver_info(task.node)

    try:
        # Restore bios config
        irmc.elcm.restore_bios_config(irmc_info, bios_config)
    except irmc.scci.SCCIError as e:
        # If the input bios config is not correct or corrupted, then
        # we should remove it from the node object.
        if isinstance(e, irmc.scci.SCCIInvalidInputError):
            _remove_bios_config(task)

        LOG.error('Failed to restore BIOS config on node %(node)s. '
                  'Error: %(error)s', {'node': node_uuid, 'error': e})
        raise exception.IRMCOperationError(operation='restore BIOS config',
                                           error=e)

    # Remove the backup data after restoring
    _remove_bios_config(task, reboot_flag=True)

    LOG.info('BIOS config is restored successfully on node %s',
             node_uuid)

    # Change power state to ON as server is automatically
    # shutdown after the operation.
    manager_utils.node_power_action(task, states.POWER_ON)


class IRMCManagement(ipmitool.IPMIManagement,
                     redfish_management.RedfishManagement):

    def get_properties(self):
        """Return the properties of the interface.

        :returns: Dictionary of <property name>:<property description> entries.
        """
        return irmc_common.COMMON_PROPERTIES

    @METRICS.timer('IRMCManagement.validate')
    def validate(self, task):
        """Validate the driver-specific management information.

        This method validates whether the 'driver_info' property of the
        supplied node contains the required information for this driver.

        :param task: A TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue if required parameters are invalid.
        :raises: MissingParameterValue if a required parameter is missing.
        """
        if task.node.driver_internal_info.get('irmc_ipmi_succeed'):
            irmc_common.parse_driver_info(task.node)
            irmc_common.update_ipmi_properties(task)
            super(IRMCManagement, self).validate(task)
        else:
            irmc_common.parse_driver_info(task.node)
            super(ipmitool.IPMIManagement, self).validate(task)

    def get_supported_boot_devices(self, task):
        """Get list of supported boot devices

        Actual code is delegated to IPMIManagement or RedfishManagement
        based on iRMC firmware version.

        :param task: A TaskManager instance
        :returns: A list with the supported boot devices defined
                  in :mod:`ironic.common.boot_devices`.

        """
        if task.node.driver_internal_info.get('irmc_ipmi_succeed'):
            return super(IRMCManagement, self).get_supported_boot_devices(task)
        else:
            return super(ipmitool.IPMIManagement,
                         self).get_supported_boot_devices(task)

    @METRICS.timer('IRMCManagement.set_boot_device')
    @task_manager.require_exclusive_lock
    def set_boot_device(self, task, device, persistent=False):
        """Set the boot device for a node.

        Set the boot device to use on next reboot of the node.

        :param task: A task from TaskManager.
        :param device: The boot device, one of the supported devices
                       listed in :mod:`ironic.common.boot_devices`.
        :param persistent: Boolean value. True if the boot device will
                           persist to all future boots, False if not.
                           Default: False.
        :raises: InvalidParameterValue if an invalid boot device is
                 specified.
        :raises: MissingParameterValue if a required parameter is missing.
        :raises: IPMIFailure on an error from ipmitool.
        :raises: RedfishConnectionError on Redfish operation failure.
        :raises: RedfishError on Redfish operation failure.
        """
        if task.node.driver_internal_info.get('irmc_ipmi_succeed'):
            if device not in self.get_supported_boot_devices(task):
                raise exception.InvalidParameterValue(_(
                    "Invalid boot device %s specified.") % device)

            uefi_mode = (
                boot_mode_utils.get_boot_mode(task.node) == 'uefi')

            # disable 60 secs timer
            timeout_disable = "0x00 0x08 0x03 0x08"
            ipmitool.send_raw(task, timeout_disable)

            # note(naohirot):
            # Set System Boot Options : ipmi cmd '0x08', bootparam '0x05'
            #
            # $ ipmitool raw 0x00 0x08 0x05 data1 data2 0x00 0x00 0x00
            #
            # data1 : '0xe0' persistent + uefi
            #         '0xc0' persistent + bios
            #         '0xa0' next only  + uefi
            #         '0x80' next only  + bios
            # data2 : boot device defined in the dict _BOOTPARAM5_DATA2

            bootparam5 = '0x00 0x08 0x05 %s %s 0x00 0x00 0x00'
            if persistent:
                data1 = '0xe0' if uefi_mode else '0xc0'
            else:
                data1 = '0xa0' if uefi_mode else '0x80'
            data2 = _BOOTPARAM5_DATA2[device]

            cmd8 = bootparam5 % (data1, data2)
            ipmitool.send_raw(task, cmd8)
        else:
            if device not in self.get_supported_boot_devices(task):
                raise exception.InvalidParameterValue(_(
                    "Invalid boot device %s specified. "
                    "Current iRMC firmware condition doesn't support IPMI "
                    "but Redfish.") % device)
            super(ipmitool.IPMIManagement, self).set_boot_device(
                task, device, persistent)

    def get_boot_device(self, task):
        """Get the current boot device for the task's node.

        Returns the current boot device of the node.

        :param task: a task from TaskManager.
        :raises: InvalidParameterValue if an invalid boot device is
                 specified.
        :raises: MissingParameterValue if a required parameter is missing.
        :raises: IPMIFailure on an error from ipmitool.
        :raises: RedfishConnectionError on Redfish operation failure.
        :raises: RedfishError on Redfish operation failure.
        :returns: a dictionary containing:

            :boot_device: the boot device, one of
                :mod:`ironic.common.boot_devices` or None if it is unknown.
            :persistent: Whether the boot device will persist to all
                future boots or not, None if it is unknown.
        """
        if task.node.driver_internal_info.get('irmc_ipmi_succeed'):
            return super(IRMCManagement, self).get_boot_device(task)
        else:
            return super(
                ipmitool.IPMIManagement, self).get_boot_device(task)

    def get_supported_boot_modes(self, task):
        """Get a list of the supported boot modes.

        IRMCManagement class doesn't support this method

        :param task: a task from TaskManager.
        :raises: UnsupportedDriverExtension if requested operation is
                 not supported by the driver
        """
        raise exception.UnsupportedDriverExtension(
            driver=task.node.driver, extension='get_supported_boot_modes')

    def set_boot_mode(self, task, mode):
        """Set the boot mode for a node.

        IRMCManagement class doesn't support this method

        :param task: a task from TaskManager.
        :param mode: The boot mode, one of
                     :mod:`ironic.common.boot_modes`.
        :raises: UnsupportedDriverExtension if requested operation is
                 not supported by the driver
        """
        raise exception.UnsupportedDriverExtension(
            driver=task.node.driver, extension='set_boot_mode')

    def get_boot_mode(self, task):
        """Get the current boot mode for a node.

        IRMCManagement class doesn't support this method

        :param task: a task from TaskManager.
        :raises: UnsupportedDriverExtension if requested operation is
                 not supported by the driver
        """
        raise exception.UnsupportedDriverExtension(
            driver=task.node.driver, extension='get_boot_mode')

    @METRICS.timer('IRMCManagement.get_sensors_data')
    def get_sensors_data(self, task):
        """Get sensors data method.

        It gets sensor data from the task's node via SCCI, and convert the data
        from XML to the dict format.

        :param task: A TaskManager instance.
        :raises: FailedToGetSensorData when getting the sensor data fails.
        :raises: FailedToParseSensorData when parsing sensor data fails.
        :raises: InvalidParameterValue if required parameters are invalid.
        :raises: MissingParameterValue if a required parameter is missing.
        :returns: Returns a consistent formatted dict of sensor data grouped
                  by sensor type, which can be processed by Ceilometer.
                  Example::

                      {
                        'Sensor Type 1': {
                          'Sensor ID 1': {
                            'Sensor Reading': 'Value1 Units1',
                            'Sensor ID': 'Sensor ID 1',
                            'Units': 'Units1'
                          },
                          'Sensor ID 2': {
                            'Sensor Reading': 'Value2 Units2',
                            'Sensor ID': 'Sensor ID 2',
                            'Units': 'Units2'
                          }
                        },
                        'Sensor Type 2': {
                          'Sensor ID 3': {
                            'Sensor Reading': 'Value3 Units3',
                            'Sensor ID': 'Sensor ID 3',
                            'Units': 'Units3'
                          },
                          'Sensor ID 4': {
                            'Sensor Reading': 'Value4 Units4',
                            'Sensor ID': 'Sensor ID 4',
                            'Units': 'Units4'
                          }
                        }
                      }

        """
        # irmc_common.parse_driver_info() makes sure that
        # d_info['irmc_sensor_method'] is either 'scci' or 'ipmitool'.
        d_info = irmc_common.parse_driver_info(task.node)
        sensor_method = d_info['irmc_sensor_method']
        if sensor_method == 'scci':
            return _get_sensors_data(task)
        elif sensor_method == 'ipmitool':
            if task.node.driver_internal_info.get('irmc_ipmi_succeed'):
                return super(IRMCManagement, self).get_sensors_data(task)
            else:
                raise exception.InvalidParameterValue(_(
                    "Invalid sensor method %s specified. "
                    "IPMI operation doesn't work on current iRMC "
                    "condition.") % sensor_method)

    @METRICS.timer('IRMCManagement.inject_nmi')
    @task_manager.require_exclusive_lock
    def inject_nmi(self, task):
        """Inject NMI, Non Maskable Interrupt.

        Inject NMI (Non Maskable Interrupt) for a node immediately.

        :param task: A TaskManager instance containing the node to act on.
        :raises: IRMCOperationError on an error from SCCI
        :returns: None

        """
        node = task.node
        irmc_client = irmc_common.get_irmc_client(node)
        try:
            irmc_client(irmc.scci.POWER_RAISE_NMI)
        except irmc.scci.SCCIClientError as err:
            LOG.error('iRMC Inject NMI failed for node %(node)s: %(err)s.',
                      {'node': node.uuid, 'err': err})
            raise exception.IRMCOperationError(
                operation=irmc.scci.POWER_RAISE_NMI, error=err)

    @METRICS.timer('IRMCManagement.restore_irmc_bios_config')
    @base.clean_step(
        priority=CONF.irmc.clean_priority_restore_irmc_bios_config)
    def restore_irmc_bios_config(self, task):
        """Restore BIOS config for a node.

        :param task: a task from TaskManager.
        :raises: NodeCleaningFailure, on failure to execute step.
        :returns: None.
        """
        try:
            _restore_bios_config(task)
        except exception.IRMCOperationError as e:
            raise exception.NodeCleaningFailure(node=task.node.uuid,
                                                reason=e)

    def get_secure_boot_state(self, task):
        """Get the current secure boot state for the node.

        NOTE: Not all drivers support this method. Older hardware
              may not implement that.

        :param task: A task from TaskManager.
        :raises: MissingParameterValue if a required parameter is missing
        :raises: DriverOperationError or its derivative in case
                 of driver runtime error.
        :raises: UnsupportedDriverExtension if secure boot is
                 not supported by the driver or the hardware
        :returns: Boolean
        """
        return irmc_common.get_secure_boot_mode(task.node)

    def set_secure_boot_state(self, task, state):
        """Set the current secure boot state for the node.

        NOTE: Not all drivers support this method. Older hardware
              may not implement that.

        :param task: A task from TaskManager.
        :param state: A new state as a boolean.
        :raises: MissingParameterValue if a required parameter is missing
        :raises: DriverOperationError or its derivative in case
                 of driver runtime error.
        :raises: UnsupportedDriverExtension if secure boot is
                 not supported by the driver or the hardware
        """
        return irmc_common.set_secure_boot_mode(task.node, state)

    def get_supported_indicators(self, task, component=None):
        """Get a map of the supported indicators (e.g. LEDs).

        IRMCManagement class doesn't support this method

        :param task: a task from TaskManager.
        :param component: If not `None`, return indicator information
            for just this component, otherwise return indicators for
            all existing components.
        :raises: UnsupportedDriverExtension if requested operation is
                 not supported by the driver

        """
        raise exception.UnsupportedDriverExtension(
            driver=task.node.driver, extension='get_supported_indicators')

    def set_indicator_state(self, task, component, indicator, state):
        """Set indicator on the hardware component to the desired state.

        IRMCManagement class doesn't support this method

        :param task: A task from TaskManager.
        :param component: The hardware component, one of
            :mod:`ironic.common.components`.
        :param indicator: Indicator ID (as reported by
            `get_supported_indicators`).
        :state: Desired state of the indicator, one of
            :mod:`ironic.common.indicator_states`.
        :raises: UnsupportedDriverExtension if requested operation is
                 not supported by the driver
        """
        raise exception.UnsupportedDriverExtension(
            driver=task.node.driver, extension='set_indicator_state')

    def get_indicator_state(self, task, component, indicator):
        """Get current state of the indicator of the hardware component.

        IRMCManagement class doesn't support this method

        :param task: A task from TaskManager.
        :param component: The hardware component, one of
            :mod:`ironic.common.components`.
        :param indicator: Indicator ID (as reported by
            `get_supported_indicators`).
        :raises: UnsupportedDriverExtension if requested operation is
                 not supported by the driver
        """
        raise exception.UnsupportedDriverExtension(
            driver=task.node.driver, extension='get_indicator_state')

    def detect_vendor(self, task):
        """Detects and returns the hardware vendor.

        :param task: A task from TaskManager.
        :raises: InvalidParameterValue if a required parameter is missing
        :raises: MissingParameterValue if a required parameter is missing
        :raises: RedfishError on Redfish operation error.
        :raises: PasswordFileFailedToCreate from creating or writing to the
                 temporary file during IPMI operation.
        :raises: processutils.ProcessExecutionError from executing ipmi command
        :returns: String representing the BMC reported Vendor or
                  Manufacturer, otherwise returns None.
        """
        if task.node.driver_internal_info.get('irmc_ipmi_succeed'):
            return super(IRMCManagement, self).detect_vendor(task)
        else:
            return super(ipmitool.IPMIManagement, self).detect_vendor(task)

    def get_mac_addresses(self, task):
        """Get MAC address information for the node.

        IRMCManagement class doesn't support this method

        :param task: A TaskManager instance containing the node to act on.
        :raises: UnsupportedDriverExtension
        """
        raise exception.UnsupportedDriverExtension(
            driver=task.node.driver, extension='get_mac_addresses')



# Copyright 2014 Rackspace, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from ironic.common import exception

LOG = log.getLogger(__name__)


def get_node_vif_ids(task):
    """Get all VIF ids for a node.

    This function does not handle multi node operations.

    :param task: a TaskManager instance.
    :returns: A dict of Node's neutron ports where keys are
        'ports' & 'portgroups' and the values are dict of UUIDs
        and their associated VIFs, e.g.

              ::

               {'ports': {'port.uuid': vif.id},
                'portgroups': {'portgroup.uuid': vif.id}}
    """
    vifs = {}
    portgroup_vifs = {}
    port_vifs = {}
    for portgroup in task.portgroups:
        vif = task.driver.network.get_current_vif(task, portgroup)
        if vif:
            portgroup_vifs[portgroup.uuid] = vif
    vifs['portgroups'] = portgroup_vifs
    for port in task.ports:
        vif = task.driver.network.get_current_vif(task, port)
        if vif:
            port_vifs[port.uuid] = vif
    vifs['ports'] = port_vifs
    return vifs


def get_portgroup_by_id(task, portgroup_id):
    """Lookup a portgroup by ID on a task object.

    :param task: a TaskManager instance
    :param portgroup_id: ID of the portgroup.
    :returns: A Portgroup object or None.
    """
    for portgroup in task.portgroups:
        if portgroup.id == portgroup_id:
            return portgroup


def get_ports_by_portgroup_id(task, portgroup_id):
    """Lookup ports by their portgroup ID on a task object.

    :param task: a TaskManager instance
    :param portgroup_id: ID of the portgroup.
    :returns: A list of Port objects.
    """
    return [port for port in task.ports if port.portgroup_id == portgroup_id]


def get_physnets_for_node(task):
    """Return the set of physical networks for a node.

    Returns the set of physical networks associated with a node's ports. The
    physical network None is excluded from the set.

    :param task: a TaskManager instance
    :returns: A set of physical networks.
    """
    return set(port.physical_network
               for port in task.ports
               if port.physical_network is not None)


def get_physnets_by_portgroup_id(task, portgroup_id, exclude_port=None):
    """Return the set of physical networks associated with a portgroup.

    :param task: a TaskManager instance.
    :param portgroup_id: ID of the portgroup.
    :param exclude_port: A Port object to exclude from the determination of the
        portgroup's physical network, or None.
    :returns: The set of physical networks associated with the portgroup. The
        set will contain zero or one physical networks.
    :raises: PortgroupPhysnetInconsistent if the portgroup's ports are not
        assigned the same physical network.
    """
    pg_ports = get_ports_by_portgroup_id(task, portgroup_id)
    if exclude_port is not None and 'id' in exclude_port:
        exclude_port_id = exclude_port.id
    else:
        exclude_port_id = None
    pg_physnets = set(port.physical_network
                      for port in pg_ports
                      if port.id != exclude_port_id)
    # Sanity check: all ports should have the same physical network.
    if len(pg_physnets) > 1:
        portgroup = get_portgroup_by_id(task, portgroup_id)
        raise exception.PortgroupPhysnetInconsistent(
            portgroup=portgroup.uuid, physical_networks=", ".join(pg_physnets))
    return pg_physnets


def remove_vifs_from_node(task):
    """Remove all vif attachment records from a node.

    :param task: a TaskManager instance.
    """
    vifs = task.driver.network.vif_list(task)
    for vif_entry in vifs:
        vif = vif_entry.get('id')
        if not vif:
            LOG.warning('Node %(node)s has an incorrect VIF entry: %(found)s. '
                        'This entry lacks an "id" field and is thus '
                        'unsupported.',
                        {'node': task.node.uuid,
                         'found': vif_entry})
            continue
        try:
            task.driver.network.vif_detach(task, vif)
        except exception.VifNotAttached:
            LOG.warning('While removing records of VIF attachments from node '
                        '%(node)s, we received indication that VIF %(vif)s is '
                        'no longer attached. This should not happen under '
                        'normal circumstances.',
                        {'node': task.node.uuid,
                         'vif': vif})

        except exception.NetworkError as e:
            LOG.error('For node %(node)s, an error occurred while removing a '
                      'VIF record for VIF %(vif)s. Error: %(error)s',
                      {'node': task.node.uuid,
                       'vif': vif,
                       'error': e})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests to assert that audit middleware works as expected.
"""

from unittest import mock

from keystonemiddleware import audit
from oslo_config import cfg

from ironic.common import exception
from ironic.tests.unit.api import base


CONF = cfg.CONF


class TestAuditMiddleware(base.BaseApiTest):
    """Provide a basic smoke test to ensure audit middleware is active.

    The tests below provide minimal confirmation that the audit middleware
    is called, and may be configured. For comprehensive tests, please consult
    the test suite in keystone audit_middleware.
    """

    @mock.patch.object(audit, 'AuditMiddleware', autospec=True)
    def test_enable_audit_request(self, mock_audit):
        CONF.set_override('enabled', True, 'audit')
        self._make_app()
        mock_audit.assert_called_once_with(
            mock.ANY,
            audit_map_file=CONF.audit.audit_map_file,
            ignore_req_list=CONF.audit.ignore_req_list)

    @mock.patch.object(audit, 'AuditMiddleware', autospec=True)
    def test_enable_audit_request_error(self, mock_audit):
        CONF.set_override('enabled', True, 'audit')
        mock_audit.side_effect = IOError("file access error")
        self.assertRaises(exception.InputFileError,
                          self._make_app)

    @mock.patch.object(audit, 'AuditMiddleware', autospec=True)
    def test_disable_audit_request(self, mock_audit):
        CONF.set_override('enabled', False, 'audit')
        self._make_app()
        self.assertFalse(mock_audit.called)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for custom SQLAlchemy types via Ironic DB."""

from oslo_db import exception as db_exc
from oslo_utils import uuidutils

import ironic.db.sqlalchemy.api as sa_api
from ironic.db.sqlalchemy import models
from ironic.tests.unit.db import base


class SqlAlchemyCustomTypesTestCase(base.DbTestCase):

    # NOTE(max_lobur): Since it's not straightforward to check this in
    #                  isolation these tests use existing db models.

    def test_JSONEncodedDict_default_value(self):
        # Create chassis w/o extra specified.
        ch1_id = uuidutils.generate_uuid()
        self.dbapi.create_chassis({'uuid': ch1_id})
        # Get chassis manually to test SA types in isolation from UOM.
        ch1 = sa_api.model_query(models.Chassis).filter_by(uuid=ch1_id).one()
        self.assertEqual({}, ch1.extra)

        # Create chassis with extra specified.
        ch2_id = uuidutils.generate_uuid()
        extra = {'foo1': 'test', 'foo2': 'other extra'}
        self.dbapi.create_chassis({'uuid': ch2_id, 'extra': extra})
        # Get chassis manually to test SA types in isolation from UOM.
        ch2 = sa_api.model_query(models.Chassis).filter_by(uuid=ch2_id).one()
        self.assertEqual(extra, ch2.extra)

    def test_JSONEncodedDict_type_check(self):
        self.assertRaises(db_exc.DBError,
                          self.dbapi.create_chassis,
                          {'extra': ['this is not a dict']})

    def test_JSONEncodedList_default_value(self):
        # Create conductor w/o extra specified.
        cdr1_id = 321321
        self.dbapi.register_conductor({'hostname': 'test_host1',
                                       'drivers': None,
                                       'id': cdr1_id})
        # Get conductor manually to test SA types in isolation from UOM.
        cdr1 = (sa_api
                .model_query(models.Conductor)
                .filter_by(id=cdr1_id)
                .one())
        self.assertEqual([], cdr1.drivers)

        # Create conductor with drivers specified.
        cdr2_id = 623623
        drivers = ['foo1', 'other driver']
        self.dbapi.register_conductor({'hostname': 'test_host2',
                                       'drivers': drivers,
                                       'id': cdr2_id})
        # Get conductor manually to test SA types in isolation from UOM.
        cdr2 = (sa_api
                .model_query(models.Conductor)
                .filter_by(id=cdr2_id)
                .one())
        self.assertEqual(drivers, cdr2.drivers)

    def test_JSONEncodedList_type_check(self):
        self.assertRaises(db_exc.DBError,
                          self.dbapi.register_conductor,
                          {'hostname': 'test_host3',
                           'drivers': {'this is not a list': 'test'}})
